#####################################################################
# Copyright 
#   John Holland <john@zoner.org>
# All rights reserved.
#
# This software is licensed as described in the file LICENSE.txt, which
# you should have received as part of this distribution.
#
######################################################################

"""
Parse a ANSI X12 data file.

Maintain context state
Start saving context and segments
Interface to read and alter segments

@todo: Attach errors to returned dicts
"""
#G{classtree X12DataNode}

#import os
#import os.path

# Intrapackage imports
import pyx12
import error_handler
import errors
import map_index
import map_if
import x12file
import path
from map_walker import walk_tree, pop_to_parent_loop  # get_pop_loops, get_push_loops


class X12DataNode(object):
    """
    Capture the segment data and X12 definition for a loop subtree
    Alter relational data
    Iterate over contents
    """
    def __init__(self, x12_node, seg_data, ntype='seg'):
        """
        """
        self.x12_map_node = x12_node
        self.type = ntype
        self.seg_data = seg_data
        self.parent = None
        self.children = []
        self.errors = []

    #{ Public Methods
    def delete(self):
        """
        Delete this node.  Mark type as deleted.
        """
        self.x12_map_node = None
        self.type = None
        self.seg_data = None
        self.parent = None
        self.children = []
        self.errors = []

    def iterate_segments(self):
        """
        Iterate over this node and children, return any segments found
        """
        raise NotImplementedError('Override in sub-class')

    def iterate_loop_segments(self):
        """
        Iterate over this node and children, return loop start and loop end
        and any segments found
        """
        raise NotImplementedError('Override in sub-class')

    def get_value(self, x12_path):
        """
        @return: the element value at the relative X12 path
        @rtype: string
        """
        raise NotImplementedError('Override in sub-class')

    def set_value(self, x12_path, val):
        """
        Set the value of simple element at the first found segment at the given path
        @param x12_path: An X12 path
        @type x12_path: string
        @param val: The new element value
        @type val: string
        """
        raise NotImplementedError('Override in sub-class')

    def exists(self, x12_path_str):
        """
        Does at least one child at the x12 path exist?
        @param x12_path_str: Relative X12 path - 2400/2430
        @type x12_path_str: string
        @return: True if found
        @rtype: boolean
        """
        (curr, new_path) = self._get_start_node(x12_path_str)
        xpath = path.X12Path(new_path)
        for n in curr._select(xpath):
            return True
        return False

    def select(self, x12_path_str):
        """
        Get a slice of sub-nodes at the relative X12 path.
        @note: All interaction/modification with a X12DataNode tree (having a loop
        root) is done in place.
        @param x12_path_str: Relative X12 path - 2400/2430
        @type x12_path_str: string
        @return: Iterator on the matching sub-nodes, relative to the instance.
        @rtype: L{node<x12context.X12DataNode>}
        """
        (curr, new_path) = self._get_start_node(x12_path_str)
        xpath = path.X12Path(new_path)
        for n in curr._select(xpath):
            if xpath.seg_id is not None:
                assert n.id == xpath.seg_id
            else:
                assert len(xpath.loop_list) > 0
                assert n.id == xpath.loop_list[-1]
            assert n.parent is not None, 'Node "%s" has no parent' % (n.id)
            yield n

    def first(self, x12_path_str):
        """
        Get the first sub-node matching the relative X12 path.
        @note: All interaction/modification with a X12DataNode tree (having a loop
        root) is done in place.
        @param x12_path_str: Relative X12 path - ie 2400/2430
        @type x12_path_str: string
        @return: The matching sub-node, relative to the instance.
        @rtype: L{node<x12context.X12DataNode>}
        """
        if not self.exists(x12_path_str):
            return None
        for node in self.select(x12_path_str):
            return node

    def count(self, x12_path_str):
        """
        Get a count of sub-nodes at the relative X12 path.
        @param x12_path_str: Relative X12 path - 2400/2430
        @type x12_path_str: string
        @return: Count of matching sub-nodes
        @rtype: int
        """
        ct = 0
        (curr, new_path) = self._get_start_node(x12_path_str)
        xpath = path.X12Path(new_path)
        for n in curr._select(xpath):
            ct += 1
        return ct

    #{ Private Methods
    def _cleanup(self):
        """
        Remove deleted nodes
        """
        self.children = [x for x in self.children if x.type is not None]

    def _get_insert_idx(self, x12_node):
        """
        Find the index of self.children before which the x12_node belongs
        Nodes will be inserted after the last node with matching ordinals
        """
        self._cleanup()
        map_idx = x12_node.pos
        idx = None
        for i in range(len(self.children)):
            if self.children[i].x12_map_node.pos <= map_idx:
                idx = i
        if idx is not None:
            return idx + 1
        return len(self.children)

    def get_first_matching_segment(self, x12_path_str):
        """
        Get first found Segment at the given relative path.  If the path is not a
        valid relative path or if the given segment index does not exist, the function
        returns None.

        @param x12_path_str: Relative X12 Path
        @type x12_path_str: string
        @return: First matching data segment
        @rtype: L{node<segment.Segment>}
        @raise X12PathError: On blank or invalid path
        """
        raise NotImplementedError('Override in sub-class')

    def _get_start_node(self, x12_path_str):
        """
        Move up the tree.  Get the new starting node and the altered path
        """
        curr = self
        while x12_path_str[:3] == '../':
            if curr.parent is None:
                raise errors.X12PathError('Current node %s does not have a parent: %s' % (self.id, x12_path_str))
            curr = curr.parent
            x12_path_str = x12_path_str[3:]
        return (curr, x12_path_str)

    def _select(self, x12path):
        """
        Get the child node at the path
        @param x12path: x12 map path
        @type x12path: L{path<path.X12Path>}
        """
        if len(x12path.loop_list) == 0:
            # Only segment left
            cur_node_id = x12path.seg_id
            qual = x12path.id_val
            for child in [x for x in self.children if x.type is not None]:
                if child.type == 'seg':
                    if child.x12_map_node.is_match_qual(child.seg_data, cur_node_id, qual):
                        yield child
                else:
                    if child.id == cur_node_id:
                        yield child
        else:
            cur_node_id = x12path.loop_list[0]
            cur_loop_list = x12path.loop_list[1:]
            for child in [x for x in self.children if x.type is not None]:
                if child.id == cur_node_id:
                    if len(cur_loop_list) == 0 and x12path.seg_id is None:
                        yield child
                    else:
                        child_path = path.X12Path(x12path.format())
                        child_path.loop_list = cur_loop_list
                        for n in child._select(child_path):
                            yield n

    def __copy__(self):
        """
        Returns a copy of this node
        """
        raise NotImplementedError('Override in sub-class')

    #{ Property Accessors
    @property
    def id(self):
        """
        @return: x12 node id
        @rtype: string
        """
        if self.x12_map_node is None:
            raise errors.EngineError('This node has been deleted')
        return self.x12_map_node.id

    @property
    def cur_path(self):
        """
        @return: x12 node path
        @rtype: string
        """
        if self.x12_map_node is None:
            raise errors.EngineError('This node has been deleted')
        return self.x12_map_node.get_path()


class X12LoopDataNode(X12DataNode):
    """
    Capture the X12 definition for a loop subtree
    Alter relational data
    Iterate over contents
    """
    def __init__(self, x12_node, end_loops=[], parent=None):
        """
        Construct an X12LoopDataNode
        """
        self.x12_map_node = x12_node
        self.type = 'loop'
        #self.seg_data = None
        self.parent = parent
        self.children = []
        self.errors = []
        self.end_loops = end_loops  # we might need to close a preceeding loop

    #{ Public Methods
    def delete(self):
        """
        Delete this node.  Mark type as deleted.
        """
        self.end_loops = []
        X12DataNode.delete(self)

    def get_value(self, x12_path_str):
        """
        Returns the element value at the given relative path.  If the path is not a
        valid relative path or if the given segment index does not exist, the function
        returns None.  If multiple values exist, this function returns the first.

        @param x12_path_str: Relative X12 Path
        @type x12_path_str: string
        @return: the element value at the relative X12 path
        @rtype: string
        @raise X12PathError: On blank or invalid path
        """
        (curr, new_path) = self._get_start_node(x12_path_str)
        seg_data = curr.get_first_matching_segment(new_path)
        if seg_data is None:
            return None
        xpath = path.X12Path(new_path)
        xpath.loop_list = []
        xpath.id_val = None
        seg_part = xpath.format()
        return seg_data.get_value(seg_part)

    def set_value(self, x12_path_str, val):
        """
        Set the value of simple element at the first found segment at the given path
        @param x12_path_str: Relative X12 Path
        @type x12_path_str: string
        @param val: The new element value
        @type val: string
        """
        (curr, new_path) = self._get_start_node(x12_path_str)
        seg_data = curr.get_first_matching_segment(new_path)
        if seg_data is None:
            raise errors.X12PathError('X12 Path is invalid or was not found: %s' % (x12_path_str))
        xpath = path.X12Path(new_path)
        xpath.loop_list = []
        xpath.id_val = None
        seg_part = xpath.format()
        seg_data.set(seg_part, val)

    def iterate_segments(self):
        """
        Iterate over this node and children
        """
        for child in [x for x in self.children if x.type is not None]:
            for a in child.iterate_segments():
                yield a

    def iterate_loop_segments(self):
        """
        Iterate over this node and children, return loop start and loop end
        """
        for loop in self.end_loops:
            yield {'node': loop, 'type': 'loop_end', 'id': loop.id}
        yield {'type': 'loop_start', 'id': self.id, 'node': self.x12_map_node}
        for child in [x for x in self.children if x.type is not None]:
            for a in child.iterate_loop_segments():
                yield a
        yield {'type': 'loop_end', 'id': self.id, 'node': self.x12_map_node}

    def add_segment(self, seg_data):
        """
        Add the segment to this loop node
        iif the segment is the anchor for a child loop, also adds the loop

        @param seg_data: Segment data
        @type seg_data: L{node<segment.Segment>} or string
        @return: New segment, or None if failed
        @rtype: L{node<x12context.X12SegmentDataNode>}
        @raise pyx12.errors.X12PathError: If invalid segment
        @todo: Check counts?
        """
        seg_data = self._get_segment(seg_data)
        x12_seg_node = self.x12_map_node.get_child_seg_node(seg_data)
        if x12_seg_node is None:
            raise errors.X12PathError('The segment %s is not a member of loop %s' %
                                      (seg_data.__repr__(), self.id))
        new_data_node = X12SegmentDataNode(x12_seg_node, seg_data, self)
        child_idx = self._get_insert_idx(x12_seg_node)
        self.children.insert(child_idx, new_data_node)
        return new_data_node

    def add_loop(self, seg_data):
        """
        Add a new loop in the correct location
        @param seg_data: Segment data
        @type seg_data: L{node<segment.Segment>} or string
        @return: New loop_data_node, or None if failed
        @rtype: L{node<x12context.X12LoopDataNode>}
        """
        seg_data = self._get_segment(seg_data)
        x12_loop_node = self.x12_map_node.get_child_loop_node(seg_data)
        if x12_loop_node is None:
            raise errors.X12PathError('The segment %s is not a member of loop %s' %
                                      (seg_data.__repr__(), self.id))
        new_data_loop = self._add_loop_node(x12_loop_node)
        # Now, add the segment
        x12_seg_node = new_data_loop.x12_map_node.get_child_seg_node(seg_data)
        new_data_node = X12SegmentDataNode(
            x12_seg_node, seg_data, new_data_loop)
        new_data_loop.add_node(new_data_node)
        return new_data_loop

    def add_node(self, data_node):
        """
        Add a X12DataNode instance
        The x12_map_node of the given data_node must be a direct child of this
        object's x12_map_node
        @param data_node: The child loop node to add
        @type data_node : L{node<x12context.X12DataNode>}
        @raise errors.X12PathError: On blank or invalid path
        """
        if data_node.x12_map_node.parent != self.x12_map_node:
            raise errors.X12PathError('The loop_data_node "%s" is not a child of "%s"' %
                                      (data_node.x12_map_node.id, self.x12_map_node.id))
        data_node.parent = self
        child_idx = self._get_insert_idx(data_node.x12_map_node)
        self.children.insert(child_idx, data_node)

    def delete_segment(self, seg_data):
        """
        Delete the given segment from this loop node
         - Do not delete the first segment in a loop
         - Does not descend into child loops
         - Only delete the first found matching segment

        @param seg_data: Segment data
        @type seg_data: L{node<segment.Segment>} or string
        @return: True if found and deleted, else False
        @rtype: Boolean
        @todo: Check counts?
        """
        seg_data = self._get_segment(seg_data)
        x12_seg_node = self.x12_map_node.get_child_seg_node(seg_data)
        if x12_seg_node is None:
            return False
            #raise errors.X12PathError, 'The segment %s is not a member of loop %s' % \
            #    (seg_data.__repr__(), self.id)
        # Iterate over data nodes, except first
        self._cleanup()
        for i in range(1, len(self.children)):
            if self.children[i].type == 'seg' and self.children[i].seg_data == seg_data:
                del self.children[i]
                return True
        return False

    def delete_node(self, x12_path_str):
        """
        Delete the first node at the given relative path.  If the path is not a
        valid relative path, return False If multiple values exist, this
        function deletes the first.

        @return: True if found and deleted, else False
        @rtype: Boolean
        @raise X12PathError: On blank or invalid path
        @todo: Check counts?
        """
        (curr, new_path) = self._get_start_node(x12_path_str)
        xpath = path.X12Path(new_path)
        for n in curr._select(xpath):
            n.delete()
            return True
        return False

    def _add_loop_node(self, x12_loop_node):
        """
        Add a loop data node to the current tree
        @param x12_loop_node: X12 Loop node
        @type x12_loop_node: L{node<map_if.loop_if>}
        @return: New X12 Loop Data Node
        @rtype: L{node<x12context.X12LoopDataNode>}
        """
        new_node = X12LoopDataNode(x12_loop_node, parent=self)
        # Iterate over data nodes
        child_idx = self._get_insert_idx(x12_loop_node)
        self.children.insert(child_idx, new_node)
        return new_node

    def get_first_matching_segment(self, x12_path_str):
        """
        Get first found Segment at the given relative path.  If the path is not a
        valid relative path or if the given segment index does not exist, the function
        returns None.

        @param x12_path_str: Relative X12 Path
        @type x12_path_str: string
        @return: First matching data segment
        @rtype: L{node<segment.Segment>}
        @raise X12PathError: On blank or invalid path
        """
        if len(x12_path_str) == 0:
            raise errors.X12PathError('Blank X12 Path')
        (curr, new_path) = self._get_start_node(x12_path_str)
        xpath = path.X12Path(new_path)
        if xpath.seg_id is None:
            return None
        if len(xpath.loop_list) == 0:
            seg_id = xpath.seg_id
            qual = xpath.id_val
            try:
                for seg in [seg for seg in curr.children if seg.type == 'seg']:
                    if seg.x12_map_node.is_match_qual(seg.seg_data, seg_id, qual):
                        return seg.seg_data
                return None
            except errors.EngineError as e:
                raise errors.X12PathError('X12 Path is invalid or was not found: %s' % (x12_path_str))
        else:
            next_id = xpath.loop_list[0]
            del xpath.loop_list[0]
            try:
                for loop in [loop for loop in curr.children if loop.type == 'loop']:
                    if loop.id == next_id:
                        return loop.get_first_matching_segment(xpath.format())
                return None
            except errors.EngineError as e:
                raise errors.X12PathError('X12 Path is invalid or was not found: %s' % (x12_path_str))

    def _get_segment(self, seg_obj):
        """
        Get a pyx12.segment.Segment instance, building one from a string
        """
        if isinstance(seg_obj, pyx12.segment.Segment):
            return seg_obj
        elif isinstance(seg_obj, str):
            (seg_term, ele_term, subele_term) = self._get_terminators()
            assert seg_term is not None, 'seg_term is none, node contains no X12SegmentDataNode children?'
            assert ele_term is not None, 'seg_term is none, node contains no X12SegmentDataNode children?'
            assert subele_term is not None, 'seg_term is none, node contains no X12SegmentDataNode children?'
            return pyx12.segment.Segment(seg_obj, seg_term, ele_term, subele_term)
        else:
            raise errors.EngineError('Unknown type %s for seg_obj %i.  Expecting a pyx12.segment.Segment or a str'
                                     % (seg_obj.__class__, seg_obj))

    def _get_terminators(self):
        for child in self.children:
            if isinstance(child, X12SegmentDataNode) and child.seg_data is not None \
                    and child.seg_data.seg_term is not None:
                return (child.seg_data.seg_term, child.seg_data.ele_term, child.seg_data.subele_term)
        return self.parent._get_terminators()

    def copy(self):
        return self.__copy__()

    def __copy__(self):
        """
        Returns a copy of this node
        """
        ret = X12LoopDataNode(self.x12_map_node)
        ret.end_loops = list(self.end_loops)
        ret.parent = self.parent
        for child in self.children:
            ret.children.append(child.copy())
        return ret


class X12SegmentDataNode(X12DataNode):
    """
    Capture the segment data and X12 definition
    Alter relational data
    Iterate over contents
    """
    def __init__(self, x12_node, seg_data, parent=None, start_loops=[],
                 end_loops=[]):
        self.x12_map_node = x12_node
        self.type = 'seg'
        self.seg_data = seg_data
        self.parent = parent
        self.start_loops = start_loops
        self.end_loops = end_loops
        self.errors = []
        self.err_isa = []
        self.err_gs = []
        self.err_st = []
        self.err_seg = []
        self.err_ele = []

    #{ Public Methods
    def handle_errh_errors(self, errh):
        """
        Attach validation errors to segment node

        @todo: move errors to parent loops if necessary
        """
        self.err_isa.extend(errh.err_isa)
        self.err_gs.extend(errh.err_gs)
        self.err_st.extend(errh.err_st)
        self.err_seg.extend(errh.err_seg)
        self.err_ele.extend(errh.err_ele)

    def delete(self):
        """
        Delete this node.  Mark type as deleted.
        """
        self.start_loops = []
        self.end_loops = []
        X12DataNode.delete(self)

    def get_value(self, x12_path_str):
        """
        Get the value of the first found element at the given path
        @param x12_path_str: Relative X12 Path
        @type x12_path_str: string
        @return: the element value at the relative X12 path
        @rtype: string
        """
        seg_data = self.get_first_matching_segment(x12_path_str)
        if seg_data is None:
            return None
        return seg_data.get_value(x12_path_str)

    def set_value(self, x12_path_str, val):
        """
        Set the value of simple element at the first found segment at the given path
        @param x12_path_str: Relative X12 Path
        @type x12_path_str: string
        @param val: The new element value
        @type val: string
        """
        seg_data = self.get_first_matching_segment(x12_path_str)
        if seg_data is None:
            raise errors.X12PathError('X12 Path is invalid or was not found: %s' % (x12_path_str))
        #ele_idx = self.get_ele_idx(x12_path_str)
        #seg_data.set(ele_idx, val)
        seg_data.set(x12_path_str, val)

    def get_first_matching_segment(self, x12_path_str):
        """
        Get first found Segment at the given relative path.  If the path is not a
        valid relative path or if the given segment index does not exist, the function
        returns None.

        @param x12_path_str: Relative X12 Path
        @type x12_path_str: string
        @return: First matching data segment
        @rtype: L{node<segment.Segment>}
        @raise X12PathError: On blank or invalid path
        """
        (curr, new_path_str) = self._get_start_node(x12_path_str)
        xpath = path.X12Path(new_path_str)
        if len(xpath.loop_list) != 0:
            raise errors.X12PathError('This X12 Path should not contain loops: %s' % (x12_path_str))
        seg_id = xpath.seg_id
        qual = xpath.id_val
        ele_idx = xpath.ele_idx
        if ele_idx is not None and seg_id is None:
            return self.seg_data
        #subele_idx = xpath.subele_idx
        try:
            if curr.x12_map_node.is_match_qual(curr.seg_data, seg_id, qual):
                return curr.seg_data
            return None
        except errors.EngineError as e:
            raise errors.X12PathError('X12 Path is invalid or was not found: %s' % (x12_path_str))
        return None

#    @staticmethod
#    def get_seg_id_parts(x12_path):
#        """
#        Split a X12 segment reference designation into component parts
#        @return: (segment ID, qualifier part, index)
#        @rtype: (string, string, string)
#        @raise X12PathError: On blank or invalid path
#        """
#        if x12_path.find('/') != -1:
#            x12_path = x12_path[x12_path.rfind('/')+1:]
#        pos = x12_path.find('[')
#        if pos != -1:
#            end = x12_path.find(']')
#            if end == -1 or end < pos:
#                raise errors.X12PathError, 'Bad X12 path: %s' % (x12_path)
#            qual = x12_path[pos+1:end]
#            seg_id = x12_path[:pos]
#            idx = x12_path[end+1:]
#            return (seg_part, qual, idx)
#        else:
#            return (x12_path, None)

    def iterate_segments(self):
        """
        Iterate on this node, return the segment
        """
        yield {'type': 'seg', 'id': self.x12_map_node.id, 'path': self.x12_map_node.x12path,
               'segment': self.seg_data}

    def iterate_loop_segments(self):
        """
        Iterate over this node and children, return loop start and loop end
        and any segments found
        """
        for loop in self.end_loops:
            yield {'node': loop, 'type': 'loop_end', 'id': loop.id}
        for loop in self.start_loops:
            yield {'node': loop, 'type': 'loop_start', 'id': loop.id}
        yield {'type': 'seg', 'id': self.id, 'segment': self.seg_data,
               'start_loops': self.start_loops, 'end_loops': self.end_loops}

    def copy(self):
        return self.__copy__()

    def __copy__(self):
        """
        Returns a copy of this node
        """
        seg_data = self.seg_data.copy()
        ret = X12SegmentDataNode(self.x12_map_node, seg_data, self.parent)
        ret.start_loops = list(self.start_loops)
        ret.end_loops = list(self.end_loops)
        return ret

    def select(self, x12_path_str):
        """
        Segment nodes have no sub-nodes so return None
        @param x12_path_str: Relative X12 path - 2400/2430
        @type x12_path_str: string
        @return: Iterator on the matching sub-nodes, relative to the instance.
        @rtype: L{node<x12context.X12DataNode>}
        """
        return []

    def _select(self, x12path):
        """
        Empty iter for segment nodes
        @param x12path: x12 map path
        @type x12path: L{path<path.X12Path>}
        """
        return []

    #{ Property Accessors
    @property
    def err_ct(self):
        """
        @return: Count of errors for this segment
        @rtype: int
        """
        return len(self.err_isa) + len(self.err_gs) + len(self.err_st) + len(self.err_seg) + len(self.err_ele)


class X12ContextReader(object):
    """
    Read an X12 input stream
    Keep context when needed
    """

    def __init__(self, param, errh, src_file_obj, xslt_files=None, map_path=None):
        """
        @param param: pyx12.param instance
        @param errh: Error Handler object
        @param src_file_obj: Source document
        @type src_file_obj: string
        @rtype: boolean
        """
        self.param = param
        self.map_path = map_path
        self.errh = error_handler.errh_list()
        self.icvn = None
        self.fic = None
        self.vriic = None
        self.tspc = None

        # Get X12 DATA file
        self.src = x12file.X12Reader(src_file_obj)

        #Get Map of Control Segments
        self.map_file = 'x12.control.00501.xml' if self.src.icvn == '00501' else 'x12.control.00401.xml'
        self.control_map = map_if.load_map_file(self.map_file, param, self.map_path)
        self.map_index_if = map_index.map_index(self.map_path)
        self.x12_map_node = self.control_map.getnodebypath('/ISA_LOOP/ISA')
        self.walker = walk_tree()

    #{ Public Methods
    def iter_segments(self, loop_id=None):
        """
        Simple segment or tree iterator
        @return: X12 Data Node - simple segment or tree
        @rtype: L{node<x12context.X12DataNode>}
        """
        cur_tree = None
        cur_data_node = None
        for seg in self.src:
            #find node
            orig_node = self.x12_map_node
            pop_loops = []
            push_loops = []
            errh = error_handler.errh_list()

            if seg.get_seg_id() == 'ISA':
                tpath = '/ISA_LOOP/ISA'
                self.x12_map_node = self.control_map.getnodebypath(tpath)
            elif seg.get_seg_id() == 'GS':
                tpath = '/ISA_LOOP/GS_LOOP/GS'
                self.x12_map_node = self.control_map.getnodebypath(tpath)
            else:
                try:
                    (seg_node, pop_loops, push_loops) = self.walker.walk(self.x12_map_node,
                            seg, errh, self.src.get_seg_count(),
                            self.src.get_cur_line(), self.src.get_ls_id())
                    self.x12_map_node = seg_node
                except errors.EngineError:
                    raise
            if self.x12_map_node is None:
                self.x12_map_node = orig_node
            else:
                seg_id = seg.get_seg_id()
                if seg_id == 'ISA':
                    icvn = seg.get_value('ISA12')
                elif seg_id == 'GS':
                    fic = seg.get_value('GS01')
                    vriic = seg.get_value('GS08')
                    map_file_new = self.map_index_if.get_filename(icvn, vriic, fic)
                    if self.map_file != map_file_new:
                        self.map_file = map_file_new
                        if self.map_file is None:
                            raise pyx12.errors.EngineError("Map not found.  icvn=%s, fic=%s, vriic=%s" %
                                                           (icvn, fic, vriic))
                        cur_map = map_if.load_map_file(self.map_file, self.param, self.map_path)
                        if cur_map.id == '837':
                            self.src.check_837_lx = True
                        else:
                            self.src.check_837_lx = False
                        #self._apply_loop_count(orig_node, cur_map)
                        #self._reset_isa_counts(cur_map)
                        self._reset_counter_to_isa_counts()
                    #self._reset_gs_counts(cur_map)
                    self._reset_counter_to_gs_counts()
                    tpath = '/ISA_LOOP/GS_LOOP/GS'
                    self.x12_map_node = cur_map.getnodebypath(tpath)
                    #self.walker.forceWalkCounterToLoopStart('/ISA_LOOP/GS_LOOP', '/ISA_LOOP/GS_LOOP/GS')
                elif seg_id == 'BHT':
                    if vriic in ('004010X094', '004010X094A1'):
                        tspc = seg.get_value('BHT02')
                        map_file_new = self.map_index_if.get_filename(icvn, vriic, fic, tspc)
                        if self.map_file != map_file_new:
                            self.map_file = map_file_new
                            if self.map_file is None:
                                err_str = "Map not found.  icvn=%s, fic=%s, vriic=%s, tspc=%s" % \
                                    (icvn, fic, vriic, tspc)
                                raise pyx12.errors.EngineError(err_str)
                            cur_map = map_if.load_map_file(self.map_file, self.param, self.map_path)
                            if cur_map.id == '837':
                                self.src.check_837_lx = True
                            else:
                                self.src.check_837_lx = False
                            self._apply_loop_count(self.x12_map_node, cur_map)
                            tpath = '/ISA_LOOP/GS_LOOP/ST_LOOP/HEADER/BHT'
                            self.x12_map_node = cur_map.getnodebypath(tpath)

            node_x12path = self.x12_map_node.x12path
            # If we are in the requested tree, wait until we have the whole thing
            if loop_id is not None and loop_id in node_x12path.loop_list:
                # Are we at the start of the requested tree?
                if node_x12path.loop_list[-1] == loop_id and \
                        self.x12_map_node.is_first_seg_in_loop():
                    if cur_tree is not None:
                        # Found root loop repeat. Yield existing, create new tree
                        yield cur_tree
                    # Make new tree on parent loop
                    #pop_loops = get_pop_loops(cur_data_node.x12_map_node, self.x12_map_node)
                    #pop_loops = [x12_node for x12_node in pop_loops if x12_node.get_path().find(loop_id) == -1]
                    cur_tree = X12LoopDataNode(x12_node=self.x12_map_node.parent, end_loops=pop_loops)  # parent=cur_data_node)
                    cur_data_node = self._add_segment(cur_tree, self.x12_map_node, seg, pop_loops, push_loops)
                else:
                    if cur_data_node is None or self.x12_map_node is None:
                        raise errors.EngineError('Either cur_data_node or self.x12_map_node is None')
                    cur_data_node = self._add_segment(cur_data_node, self.x12_map_node, seg, pop_loops, push_loops)
            else:
                if cur_tree is not None:
                    # We have completed a tree
                    yield cur_tree
                    cur_tree = None
                if cur_data_node is not None:
                    #push_loops = get_push_loops(cur_data_node.x12_map_node, self.x12_map_node)
                    #pop_loops = get_pop_loops(cur_data_node.x12_map_node, self.x12_map_node)
                    if loop_id:
                        pop_loops = [x12_node for x12_node in pop_loops if x12_node.get_path().find(loop_id) == -1]
                    assert loop_id not in [x12.id for x12 in push_loops], 'Loop ID %s should not be in push loops' % (loop_id)
                    assert loop_id not in [x12.id for x12 in pop_loops], 'Loop ID %s should not be in pop loops' % (loop_id)
                    cur_data_node = X12SegmentDataNode(self.x12_map_node, seg, push_loops, pop_loops)
                else:
                    cur_data_node = X12SegmentDataNode(self.x12_map_node, seg)
                # Get errors caught by x12Reader
                errh.handle_errors(self.src.pop_errors())
                # Handle errors captured in errh_list
                cur_data_node.handle_errh_errors(errh)
                if cur_data_node.id != 'ISA' and cur_data_node is not None:
                    assert cur_data_node.parent is not None, 'Node "%s" has no parent' % (cur_data_node.id)
                yield cur_data_node

    def register_error_callback(self, callback, err_type):
        """
        Future:  Callbacks for X12 validation errors
        """
        pass

    #{ Property Accessors
    @property
    def seg_term(self):
        """
        @return: Current X12 segment terminator
        @rtype: string
        """
        return self.src.seg_term

    @property
    def ele_term(self):
        """
        @return: Current X12 element terminator
        @rtype: string
        """
        return self.src.ele_term

    @property
    def subele_term(self):
        """
        @return: Current X12 sub-element terminator
        @rtype: string
        """
        return self.src.subele_term

    #{ Private Methods
    def _add_segment(self, cur_data_node, segment_x12_node, seg_data, pop_loops, push_loops):
        """
        From the last position in the X12 Data Node Tree, find the correct
        position for the new segment; moving up or down the tree as appropriate.

        G{callgraph}

        @param cur_data_node: Current X12 Data Node
        @type cur_data_node: L{node<x12context.X12DataNode>}
        @param segment_x12_node: Segment Map Node
        @type segment_x12_node: L{node<map_if.x12_node>}
        @return: New X12 Data Node
        @rtype: L{node<x12context.X12DataNode>}
        """
        if not segment_x12_node.is_segment():
            raise errors.EngineError('Node must be a segment')
        # Get enclosing loop
        orig_data_node = cur_data_node
        parent_x12_node = pop_to_parent_loop(segment_x12_node)
        cur_loop_node = cur_data_node
        if cur_loop_node.type == 'seg':
            cur_loop_node = cur_loop_node.parent
        # check path for new loops to be added
        new_path = parent_x12_node.x12path
        last_path = cur_loop_node.x12_map_node.x12path
        if last_path != new_path:
            for x12_loop in pop_loops:
                if cur_loop_node.id != x12_loop.id:
                    raise errors.EngineError('Loop pop: %s != %s' %
                                             (cur_loop_node.id, x12_loop.id))
                cur_loop_node = cur_loop_node.parent
            for x12_loop in push_loops:
                if cur_loop_node is None:
                    raise errors.EngineError('cur_loop_node is None. x12_loop: %s' % (x12_loop.id))
                # push new loop nodes, if needed
                cur_loop_node = cur_loop_node._add_loop_node(x12_loop)
        else:
            # handle loop repeat
            if cur_loop_node.parent is not None and segment_x12_node.is_first_seg_in_loop():
                cur_loop_node = cur_loop_node.parent._add_loop_node(
                    segment_x12_node.parent)
        try:
            new_node = X12SegmentDataNode(self.x12_map_node, seg_data)
        except Exception:
            mypath = self.x12_map_node.get_path()
            err_str = 'X12SegmentDataNode failed: x12_path={}, seg_date={}'.format(mypath, seg_data)
            raise errors.EngineError(err_str)
        try:
            new_node.parent = cur_loop_node
            cur_loop_node.children.append(new_node)
        except Exception:
            err_str = 'X12SegmentDataNode child append failed:'
            err_str += ' seg_x12_path=%s' % (segment_x12_node.get_path())
            err_str += ', orig_datanode=%s' % (orig_data_node.cur_path)
            err_str += ', cur_datanode=%s' % (cur_data_node.cur_path)
            err_str += ', seg_data=%s' % (seg_data)
            raise errors.EngineError(err_str)
        return new_node

    #def _apply_loop_count(self, orig_node, new_map):
    #    """
    #    Apply loop counts to current map
    #    """
    #    ct_list = []
    #    orig_node.get_counts_list(ct_list)
    #    for (path1, ct) in ct_list:
    #        curnode = new_map.getnodebypath(path1)
    #        curnode.set_cur_count(ct)

    #def _reset_isa_counts(self, cur_map):
    #    """
    #    Reset ISA instance counts
    #    """
    #    cur_map.getnodebypath('/ISA_LOOP').set_cur_count(1)
    #    cur_map.getnodebypath('/ISA_LOOP/ISA').set_cur_count(1)

    #def _reset_gs_counts(self, cur_map):
    #    """
    #    Reset GS instance counts
    #    """
    #    cur_map.getnodebypath('/ISA_LOOP/GS_LOOP').reset_cur_count()
    #    cur_map.getnodebypath('/ISA_LOOP/GS_LOOP').set_cur_count(1)
    #    cur_map.getnodebypath('/ISA_LOOP/GS_LOOP/GS').set_cur_count(1)

    def _reset_counter_to_isa_counts(self):
        """
        Reset ISA instance counts
        """
        self.walker.counter.reset_to_node('/ISA_LOOP')
        self.walker.counter.increment('/ISA_LOOP')
        self.walker.counter.increment('/ISA_LOOP/ISA')

    def _reset_counter_to_gs_counts(self):
        """
        Reset GS instance counts
        """
        self.walker.counter.reset_to_node('/ISA_LOOP/GS_LOOP')
        self.walker.counter.increment('/ISA_LOOP/GS_LOOP')
        self.walker.counter.increment('/ISA_LOOP/GS_LOOP/GS')



######################################################################
# Copyright Kalamazoo Community Mental Health Services,
#   John Holland <jholland@kazoocmh.org> <john@zoner.org>
# All rights reserved.
#
# This software is licensed as described in the file LICENSE.txt, which
# you should have received as part of this distribution.
#
######################################################################

"""
Base Interface to X12 Errors
"""

import abc

logger = logging.getLogger('pyx12.error_handler')


class ErrorHandlerBase(object):
    """
    Base class for the error handling structures.
    """
    __metaclass__  = abc.ABCMeta

    @abc.abstractmethod
    def accept(self, visitor):
        """
        Params:     visitor - ref to visitor class
        """

    @abc.abstractmethod
    def handle_errors(self, err_list):
        """
        @param err_list: list of errors to apply
        """

    @abc.abstractmethod
    def get_cur_line(self):
        """
        @return: Current file line number
        @rtype: int
        """

    @abc.abstractmethod
    def get_id(self):
        """
        """

    @abc.abstractmethod
    def add_isa_loop(self, seg_data, src):
        """
        @param seg_data: Segment object
        @type seg_data: L{segment<segment.Segment>}
        """

    @abc.abstractmethod
    def add_gs_loop(self, seg_data, src):
        """
        @param seg_data: Segment object
        @type seg_data: L{segment<segment.Segment>}
        """

    @abc.abstractmethod
    def add_st_loop(self, seg_data, src):
        """
        @param seg_data: Segment object
        @type seg_data: L{segment<segment.Segment>}
        """

    @abc.abstractmethod
    def add_seg(self, map_node, seg_data, seg_count, cur_line, ls_id):
        """
        @param map_node: current segment node
        @type map_node: L{node<map_if.segment_if>}
        @param seg_data: Segment object
        @type seg_data: L{segment<segment.Segment>}
        @param seg_count: Count of current segment in the ST Loop
        @type seg_count: int
        @param cur_line: Current line number in the file
        @type cur_line: int
        @param ls_id: The current LS loop identifier
        @type ls_id: string
        """

    @abc.abstractmethod
    def _add_cur_seg(self):
        """
        """

    @abc.abstractmethod
    def add_ele(self, map_node):
        """
        """

    @abc.abstractmethod
    def _add_cur_ele(self):
        """
        """

    @abc.abstractmethod
    def isa_error(self, err_cde, err_str):
        """
        @param err_cde: ISA level error code
        @type err_cde: string
        @param err_str: Description of the error
        @type err_str: string
        """

    @abc.abstractmethod
    def gs_error(self, err_cde, err_str):
        """
        @param err_cde: GS level error code
        @type err_cde: string
        @param err_str: Description of the error
        @type err_str: string
        """

    @abc.abstractmethod
    def st_error(self, err_cde, err_str):
        """
        @param err_cde: Segment level error code
        @type err_cde: string
        @param err_str: Description of the error
        @type err_str: string
        """

    @abc.abstractmethod
    def seg_error(self, err_cde, err_str, err_value=None, src_line=None):
        """
        @param err_cde: Segment level error code
        @type err_cde: string
        @param err_str: Description of the error
        @type err_str: string
        """

    @abc.abstractmethod
    def ele_error(self, err_cde, err_str, bad_value, refdes=None):
        """
        @param err_cde: Element level error code
        @type err_cde: string
        @param err_str: Description of the error
        @type err_str: string
        """

    @abc.abstractmethod
    def close_isa_loop(self, node, seg, src):

    @abc.abstractmethod
    def close_gs_loop(self, node, seg, src):

    @abc.abstractmethod
    def close_st_loop(self, node, seg, src):

    @abc.abstractmethod
    def find_node(self, type):
        """
        Find the last node of a type
        """

    @abc.abstractmethod
    def _get_last_child(self):

    @abc.abstractmethod

    @abc.abstractmethod
    def get_error_count(self):

    @abc.abstractmethod
    def get_first_child(self):

    @abc.abstractmethod
    def get_next_sibling(self):

    @abc.abstractmethod
    def __next__(self):
        """
        Return the next error node
        """

    @abc.abstractmethod
    def is_closed(self):
        """
        @rtype: boolean
        """

    @abc.abstractmethod
    def __repr__(self):
        """
        """

    @abc.abstractmethod
    def pop_errors(self):
        """
        Pop error list
        @return: List of errors
        """



######################################################################
# Copyright (c)
#   John Holland <john@zoner.org>
# All rights reserved.
#
# This software is licensed as described in the file LICENSE.txt, which
# you should have received as part of this distribution.
#
######################################################################

"""
Holds Run-time Parameters

Order of precedence:
 1. set(param) - Command line parameters
 2. Config files as constructor parameters
 3. self.params - Defaults
"""
from os.path import dirname, abspath, join, isdir, isfile, expanduser
import sys
import xml.etree.cElementTree as et
import logging

from pyx12.errors import EngineError


class ParamsBase(object):
    """
    Base class for parameters
    """
    def __init__(self):
        self.logger = logging.getLogger('pyx12.params')
        self.params = {}
        #First, try relative path
        base_dir = dirname(dirname(abspath(sys.argv[0])))
        map_path = join(base_dir, 'map')
        #Then look in standard installation location
        if not isdir(map_path):
            map_path = join(sys.prefix, 'share', 'pyx12', 'map')
        self.params['map_path'] = map_path
        self.params['exclude_external_codes'] = None
        self.params['charset'] = 'E'
        self.params['simple_dtd'] = ''
        self.params['xmlout'] = 'simple'

    def get(self, option):
        """
        Get the value of the parameter specified by option
        @param option: Option name
        @type option: string
        """
        if option in list(self.params.keys()):
            return self.params[option]
        else:
            return None

    def set(self, option, value):
        """
        Set the value of the parameter specified by option
        @param option: Option name
        @type option: string
        @param value: Parameter value
        @type value: string
        """
        if value == '':
            self.params[option] = None
        else:
            self.params[option] = value

    def _read_config_file(self, filename):
        """
        Read program configuration from an XML file

        @param filename: XML file
        @type filename: string
        @raise EngineError: If the config file is not found or is unreadable
        @return: None
        """
        if not isfile(filename):
            self.logger.debug('Configuration file "%s" does not exist' %
                              filename)
            raise EngineError('Configuration file "%s" does not exist' %
                              (filename))
        try:
            self.logger.debug('parsing config file %s' % (filename))
            t = et.parse(filename)
            for c in t.iter('param'):
                option = c.get('name')
                value = c.findtext('value')
                valtype = c.findtext('type')
                self._set_option(option, value, valtype)
        except Exception:
            self.logger.error('Read of configuration file "%s" failed' %
                              (filename))
            raise

    def _set_option(self, option, value, valtype):
        """
        Set the value of the parameter specified by option
        @param option: Option name
        @type option: string
        @param value: Parameter value
        @type value: string
        @param valtype: Parameter type
        @type valtype: string
        """
        if option is None or option == '':
            return
        if value == '':
            value = None
        if valtype == 'boolean':
            if value in ('False', 'F'):
                self.params[option] = False
            else:
                self.params[option] = True
        else:
            try:
                if self.params[option] != value:
                    self.params[option] = value
                    #self.logger.debug('Params: option "%s": "%s"' % \
                    #    (option, self.params[option]))
            except Exception:
                self.params[option] = value
                #self.logger.debug('Params: option "%s": "%s"' % \
                #   (option, self.params[option]))
        #self.logger.debug('Params: option "%s": "%s"' % \
        #    (option, self.params[option]))


class ParamsUnix(ParamsBase):
    """
    Read options from XML configuration files
    """
    def __init__(self, config_file=None):
        ParamsBase.__init__(self)
        config_files = [join(sys.prefix, 'etc/pyx12.conf.xml'),
                        expanduser('~/.pyx12.conf.xml')]
        for filename in config_files:
            if isfile(filename):
                self.logger.debug('Read param file: %s' % (filename))
                self._read_config_file(filename)
        if config_file:
            self.logger.debug('Read param file: %s' % (filename))
            self._read_config_file(config_file)
        else:
            self.logger.debug('No config file passed to the constructor')


class ParamsWindows(ParamsBase):
    """
    Read options from XML configuration files
    """
    def __init__(self, config_file=None):
        ParamsBase.__init__(self)
        config_files = [join(sys.prefix, 'etc/pyx12.conf.xml')]
        for filename in config_files:
            if isfile(filename):
                self.logger.debug('Read param file: %s' % (filename))
                self._read_config_file(filename)
        if config_file:
            self.logger.debug('Read param file: %s' % (filename))
            self._read_config_file(config_file)

if sys.platform == 'win32':
    params = ParamsWindows
else:
    params = ParamsUnix



######################################################################
# Copyright 
#   John Holland <john@zoner.org>
# All rights reserved.
#
# This software is licensed as described in the file LICENSE.txt, which
# you should have received as part of this distribution.
#
######################################################################

"""
Interface to an X12 data stream.
 - Efficiently handles large files.
 - Tracks end of explicit loops.
 - Tracks segment/line/loop counts.
 - Tracks some transaction specific counters:
   837 2400/LX
   837 HL tree
"""

import codecs
import sys
import logging

# Intrapackage imports
import pyx12.errors
import pyx12.segment
from pyx12.rawx12file import RawX12File

logger = logging.getLogger('pyx12.x12file')


class X12Base(object):
    """
    Base class of X12 Reader and X12 Writer
    Common X12 validation
    """

    def __init__(self):
        """
        Initialize the X12 file
        """
        self.err_list = []
        self.loops = []
        self.hl_stack = []
        self.gs_count = 0
        self.st_count = 0
        self.hl_count = 0
        self.seg_count = 0
        self.cur_line = 0
        self.isa_ids = []
        self.gs_ids = []
        self.st_ids = []
        self.lx_count = 0
        self.check_837_lx = False
        self.isa_usage = None
        self.seg_term = None
        self.ele_term = None
        self.subele_term = None
        self.repetition_term = None

    def Close(self):
        """
        Complete any outstanding tasks
        """
        pass

    def _parse_segment(self, seg_data):
        """
        Catch segment issues common to both readers and writers

        @param seg_data: Segment data instance
        @type seg_data: L{segment<segment.Segment>}
        """
        if seg_data.is_empty():
            err_str = 'Segment "%s" is empty' % (seg_data)
            self._seg_error('8', err_str, None, src_line=self.cur_line + 1)
        if not seg_data.is_seg_id_valid():
            err_str = 'Segment identifier "%s" is invalid' % (
                seg_data.get_seg_id())
            self._seg_error('1', err_str, None, src_line=self.cur_line + 1)
        seg_id = seg_data.get_seg_id()
        if seg_id == 'ISA':
            if len(seg_data) != 16:
                raise pyx12.errors.X12Error('The ISA segment must have 16 elements (%s)' % (seg_data))
            interchange_control_number = seg_data.get_value('ISA13')
            if interchange_control_number in self.isa_ids:
                err_str = 'ISA Interchange Control Number '
                err_str += '%s not unique within file' % (interchange_control_number)
                self._isa_error('025', err_str)
            self.loops.append(('ISA', interchange_control_number))
            self.isa_ids.append(interchange_control_number)
            self.gs_count = 0
            self.gs_ids = []
            self.isa_usage = seg_data.get_value('ISA15')
        elif seg_id == 'GS':
            group_control_number = seg_data.get_value('GS06')
            if group_control_number in self.gs_ids:
                err_str = 'GS Interchange Control Number '
                err_str += '%s not unique within file' % (group_control_number)
                self._gs_error('6', err_str)
            self.gs_count += 1
            self.gs_ids.append(group_control_number)
            self.loops.append(('GS', group_control_number))
            self.st_count = 0
            self.st_ids = []
        elif seg_id == 'ST':
            self.hl_stack = []
            self.hl_count = 0
            transaction_control_number = seg_data.get_value('ST02')
            if transaction_control_number in self.st_ids:
                err_str = 'ST Interchange Control Number '
                err_str += '%s not unique within file' % (transaction_control_number)
                self._st_error('23', err_str)
            self.st_count += 1
            self.st_ids.append(transaction_control_number)
            self.loops.append(('ST', transaction_control_number))
            self.seg_count = 1
            self.hl_count = 0
        #elif seg_id == 'LS':
        #    self.seg_count += 1
        #    self.loops.append(('LS', seg_data.get_value('LS06')))
        #elif seg_id == 'LE':
        #    self.seg_count += 1
        #    del self.loops[-1]
        elif seg_id == 'HL':
            self.hl_count += 1
            hl_count = seg_data.get_value('HL01')
            if self.hl_count != self._int(hl_count):
                #raise pyx12.errors.X12Error, \
                #   'My HL count %i does not match your HL count %s' \
                #    % (self.hl_count, seg[1])
                err_str = 'My HL count %i does not match your HL count %s' % (self.hl_count, hl_count)
                self._seg_error('HL1', err_str)
            if seg_data.get_value('HL02') != '':
                hl_parent = self._int(seg_data.get_value('HL02'))
                if hl_parent not in self.hl_stack:
                    err_str = 'HL parent (%i) is not a valid parent' % (hl_parent)
                    self._seg_error('HL2', err_str)
                while self.hl_stack and hl_parent != self.hl_stack[-1]:
                    del self.hl_stack[-1]
            else:
                if len(self.hl_stack) != 0:
                    pass
                    #err_str = 'HL parent is blank, but stack not empty'
                    #self._seg_error('HL2', err_str)
            self.hl_stack.append(self.hl_count)
        elif self.check_837_lx and seg_id == 'CLM':
            self.lx_count = 0
        elif self.check_837_lx and seg_id == 'LX':
            self.lx_count += 1
            if seg_data.get_value('LX01') != '%i' % (self.lx_count):
                err_str = 'Your 2400/LX01 Service Line Number %s does not match my count of %i' % \
                    (seg_data.get_value('LX01'), self.lx_count)
                self._seg_error('LX', err_str)
        # count all regular segments
        if seg_id not in ('ISA', 'IEA', 'GS', 'GE', 'ST', 'SE'):
            self.seg_count += 1
        self.cur_line += 1

    def pop_errors(self):
        """
        Pop error list
        @return: List of errors
        """
        tmp = self.err_list
        self.err_list = []
        return tmp

    def _isa_error(self, err_cde, err_str):
        """
        @param err_cde: ISA level error code
        @type err_cde: string
        @param err_str: Description of the error
        @type err_str: string
        """
        self.err_list.append(('isa', err_cde, err_str, None, None))

    def _gs_error(self, err_cde, err_str):
        """
        @param err_cde: GS level error code
        @type err_cde: string
        @param err_str: Description of the error
        @type err_str: string
        """
        self.err_list.append(('gs', err_cde, err_str, None, None))

    def _st_error(self, err_cde, err_str):
        """
        @param err_cde: Segment level error code
        @type err_cde: string
        @param err_str: Description of the error
        @type err_str: string
        """
        self.err_list.append(('st', err_cde, err_str, None, None))

    def _seg_error(self, err_cde, err_str, err_value=None, src_line=None):
        """
        @param err_cde: Segment level error code
        @type err_cde: string
        @param err_str: Description of the error
        @type err_str: string
        """
        self.err_list.append(('seg', err_cde, err_str, err_value, src_line))

    def _int(self, str_val):
        """
        Converts a string to an integer
        @type str_val: string
        @return: Int value if successful, None if not
        @rtype: int
        """
        try:
            return int(str_val)
        except ValueError:
            return None
        return None

    def get_isa_id(self):
        """
        Get the current ISA identifier

        @rtype: string
        """
        for loop in self.loops:
            if loop[0] == 'ISA':
                return loop[1]
        return None

    def get_gs_id(self):
        """
        Get the current GS identifier

        @rtype: string
        """
        for loop in self.loops:
            if loop[0] == 'GS':
                return loop[1]
        return None

    def get_st_id(self):
        """
        Get the current ST identifier

        @rtype: string
        """
        for loop in self.loops:
            if loop[0] == 'ST':
                return loop[1]
        return None

    def get_ls_id(self):
        """
        Get the current LS identifier

        @rtype: string
        """
        for loop in self.loops:
            if loop[0] == 'LS':
                return loop[1]
        return None

    def get_seg_count(self):
        """
        Get the current segment count

        @rtype: int
        """
        return self.seg_count

    def get_cur_line(self):
        """
        Get the current line

        @rtype: int
        """
        return self.cur_line

    def get_term(self):
        """
        Get the original terminators

        @rtype: tuple(string, string, string, string)
        """
        return (self.seg_term, self.ele_term, self.subele_term, '\n', self.repetition_term)


class X12Reader(X12Base):
    """
    Read an X12 data file

    Errors found when reading the segment such as loop counting or ID
    errors can be retrieved using the pop_errors function
    """

    def __init__(self, src_file_obj):
        """
        Initialize the file X12 file reader

        @param src_file_obj: absolute path of source file or an open,
            readable file object
        @type src_file_obj: string or open file object
        """
        self.fd_in = None
        self.need_to_close = False
        try:
            res = src_file_obj.closed
            self.fd_in = src_file_obj
        except AttributeError:
            if src_file_obj == '-':
                self.fd_in = sys.stdin
            else:
                self.fd_in = file(src_file_obj, 'U')
                self.need_to_close = True
        X12Base.__init__(self)
        try:
            self.raw = RawX12File(self.fd_in)
        except pyx12.errors.X12Error:
            raise
        (seg_term, ele_term, subele_term, eol, repetition_term) = self.raw.get_term()
        self.seg_term = seg_term
        self.ele_term = ele_term
        self.subele_term = subele_term
        self.repetition_term = repetition_term
        self.icvn = self.raw.icvn

    def __del__(self):
        try:
            if self.need_to_close:
                self.fd_in.close()
        except Exception:
            pass

    def _parse_segment(self, seg_data):
        """
        Catch segment issues

        @param seg_data: Segment data instance
        @type seg_data: L{segment<segment.Segment>}
        """
        X12Base._parse_segment(self, seg_data)
        seg_id = seg_data.get_seg_id()
        if seg_id == 'IEA':
            if self.loops[-1][0] != 'ISA':
                # Unterminated GS loop
                err_str = 'Unterminated Loop %s' % (self.loops[-1][0])
                self._isa_error('024', err_str)
                del self.loops[-1]
            if self.loops[-1][1] != seg_data.get_value('IEA02'):
                err_str = 'IEA id=%s does not match ISA id=%s' % \
                    (seg_data.get_value('IEA02'), self.loops[-1][1])
                self._isa_error('001', err_str)
            if self._int(seg_data.get_value('IEA01')) != self.gs_count:
                err_str = 'IEA count for IEA02=%s is wrong' % \
                    (seg_data.get_value('IEA02'))
                self._isa_error('021', err_str)
            del self.loops[-1]
        elif seg_id == 'GE':
            if self.loops[-1][0] != 'GS':
                err_str = 'Unterminated segment %s' % (self.loops[-1][1])
                self._gs_error('3', err_str)
                del self.loops[-1]
            if self.loops[-1][1] != seg_data.get_value('GE02'):
                err_str = 'GE id=%s does not match GS id=%s' % \
                    (seg_data.get_value('GE02'), self.loops[-1][1])
                self._gs_error('4', err_str)
            if self._int(seg_data.get_value('GE01')) != self.st_count:
                err_str = 'GE count of %s for GE02=%s is wrong. I count %i'\
                    % (seg_data.get_value('GE01'),
                       seg_data.get_value('GE02'), self.st_count)
                self._gs_error('5', err_str)
            del self.loops[-1]
        elif seg_id == 'SE':
            se_trn_control_num = seg_data.get_value('SE02')
            if self.loops[-1][0] != 'ST' or \
                    self.loops[-1][1] != se_trn_control_num:
                err_str = 'SE id=%s does not match ST id=%s' % \
                    (se_trn_control_num, self.loops[-1][1])
                self._st_error('3', err_str)
            if self._int(seg_data.get_value('SE01')) != self.seg_count + 1:
                err_str = 'SE count of %s for SE02=%s is wrong. I count %i'\
                    % (seg_data.get_value('SE01'),
                        se_trn_control_num, self.seg_count + 1)
                self._st_error('4', err_str)
            del self.loops[-1]

    def __iter__(self):
        """
        Iterate over input segments
        """
        self.err_list = []
        for line in self.raw:
            # We have not yet incremented cur_line
            if line.startswith(' ') :
                err_str = 'Segment contains a leading space'
                self._seg_error('1', err_str, None, src_line=self.cur_line + 1)
                line = line.lstrip()
            if line[-1] == self.ele_term:
                err_str = 'Segment contains trailing element terminators'
                self._seg_error('SEG1', err_str, None, src_line=self.cur_line + 1)
            seg_data = pyx12.segment.Segment(line, self.seg_term, self.ele_term, self.subele_term)
            self._parse_segment(seg_data)
            yield(seg_data)
        #yield(None)

    def cleanup(self):
        """
        At EOF, check for missing loop trailers
        """
        if self.loops:
            for (seg, id1) in self.loops:
                if seg == 'ST':
                    err_str = 'Mandatory segment "Transaction Set Trailer" '
                    err_str += '(SE=%s) missing' % (id1)
                    self._st_error('2', err_str)
                elif seg == 'GS':
                    err_str = 'Mandatory segment "Functional Group Trailer" '
                    err_str += '(GE=%s) missing' % (id1)
                    self._gs_error('3', err_str)
                elif seg == 'ISA':
                    err_str = 'Mandatory segment "Interchange Control Trailer" '
                    err_str += '(IEA=%s) missing' % (id1)
                    self._isa_error('023', err_str)
                #elif self.loops[-1][0] == 'LS':
                #    err_str = 'LS id=%s was not closed with a LE' % \
                #    (id1, self.loops[-1][1])

# Backward compatible name
X12file = X12Reader


class X12Writer(X12Base):
    """
    X12 file and stream writer
    """

    def __init__(self, src_file_obj, seg_term='~', ele_term='*', subele_term='\\', eol='\n', repetition_term='^'):
        """
        Initialize the file X12 file writer

        @param src_file_obj: absolute path of source file or an open,
            readable file object
        @type src_file_obj: string or open file object
        """
        self.fd_out = None
        try:
            res = src_file_obj.write
            # isinstance(f, file)
            self.fd_out = src_file_obj
        except AttributeError:
            if src_file_obj == '-':
                self.fd_out = sys.stdout
            else:
                self.fd_out = codecs.open(
                    src_file_obj, mode='w', encoding='ascii')
        #assert self.fd_out.encoding in ('ascii', 'US-ASCII'), 'Outfile file must have ASCII encoding, is %s' % (self.fd_out.encoding)
        X12Base.__init__(self)
        #terms = set([seg_term, ele_term, subele_term, repetition_term])
        self.seg_term = seg_term
        self.ele_term = ele_term
        self.subele_term = subele_term
        self.repetition_term = repetition_term
        self.eol = eol

    def Close(self):
        """
        End any open loops.  Should be called at the end of writing.
        """
        self._popToLoop('ISA')
        X12Base.Close(self)

    def Write(self, seg_data):
        """
        Write the segment to the stream given current separators

        @param seg_data: Segment data instance
        @type seg_data: L{segment<segment.Segment>}
        """
        self._parse_segment(seg_data)
        # If we have hit a loop closing segment, generate any missing, containing, closing segments
        # then generate this segment
        seg_id = seg_data.get_seg_id()
        if seg_id == 'IEA':
            self._popToLoop('ISA')
        elif seg_id == 'GE':
            self._popToLoop('GS')
        elif seg_id == 'SE':
            self._popToLoop('ST')
        elif self.check_837_lx and seg_id == 'LX':
            # Write our own LX counter
            seg_data.set('01', '%i' % (self.lx_count))
            self._write_segment(seg_data)
        elif seg_id == 'ISA':
            # Replace terminators
            self._write_isa_segment(seg_data)
        else:
            self._write_segment(seg_data)

    def _close_loop(self, loop_type, loop_id):
        if loop_type == 'ISA':
            self._close_iea(loop_id)
        elif loop_type == 'GS':
            self._close_ge(loop_id)
        elif loop_type == 'ST':
            self._close_se(loop_id)

    def _popToLoop(self, loop_type):
        """
        Move up the loop open loops, up to and including the given loop

        @param loop_type: The current ending loop
        @type loop_type: string
        """
        while len(self.loops) > 0 and self.loops[-1][0] != loop_type:
            loop = self.loops.pop()
            self._close_loop(loop[0], loop[1])
        if len(self.loops) > 0:
            loop = self.loops.pop()
            self._close_loop(loop[0], loop[1])

    def _close_iea(self, id):
        """
        Close a ISA/IEA loop, reset GS counter

        @param id: ISA loop ID
        @type id: string
        """
        seg_temp = self._get_trailer_segment('IEA', self.gs_count, id)
        self._write_segment(seg_temp)
        self.gs_count = 0

    def _close_ge(self, id):
        """
        Close a GS/GE loop, reset ST counter

        @param id: GS loop ID
        @type id: string
        """
        seg_temp = self._get_trailer_segment('GE', self.st_count, id)
        self._write_segment(seg_temp)
        self.st_count = 0

    def _close_se(self, id):
        """
        Close a ST/SE loop, reset segment counter

        @param id: ST loop ID
        @type id: string
        """
        seg_temp = self._get_trailer_segment('SE', self.seg_count + 1, id)
        self._write_segment(seg_temp)
        self.seg_count = 0

    def _write_segment(self, seg_data):
        """
        Write the given segment, using the current delimiters and end of line

        @param seg_data: segment to write
        @type seg_data: L{segment<segment.Segment>}
        """
        out = seg_data.format(self.seg_term, self.ele_term, self.subele_term) + self.eol
        self.fd_out.write(out.decode('ascii'))

    def _write_isa_segment(self, seg_data):
        """
        Write the ISA segment, using the current delimiters and end of line

        ISA*03*SENDER    *01*          *ZZ*SENDER         *ZZ*RECEIVER       *040608*1333*U*00401*000000288*0*P*:~
        ISA*03*SENDER    *01*          *ZZ*SENDER         *ZZ*RECEIVER       *040611*1333*^*00501*000000125*0*P*\~

        @param seg_data: ISA segment to write
        @type seg_data: L{segment<segment.Segment>}
        """
        icvn = seg_data.get_value('ISA12')
        if icvn == '00501':
            seg_data.set('ISA11', self.repetition_term)
        seg_data.set('ISA16', self.subele_term)
        out = seg_data.format(
            self.seg_term, self.ele_term, self.subele_term) + self.eol
        self.fd_out.write(out.decode('ascii'))

    def _get_trailer_segment(self, seg_id, count, id):
        """
        Create a loop trailer segment, using the matching loop start and current count

        @param seg_id: end loop segment id
        @type seg_id: string
        @param count: count of loop members
        @type count: non-negative int
        @param id: loop id, should come from loop header
        @type id: string
        """
        ele_term = self.ele_term
        seg_str = '%s%s%i%s%s' % (seg_id, ele_term, count, ele_term, id)
        return pyx12.segment.Segment(seg_str, self.seg_term, self.ele_term,
                                     self.subele_term)



######################################################################
# Copyright 
#   John Holland <john@zoner.org>
# All rights reserved.
#
# This software is licensed as described in the file LICENSE.txt, which
# you should have received as part of this distribution.
#
######################################################################

"""
Create an X12 document from a XML data file in the simple form
"""

import xml.etree.cElementTree as et
import logging

# Intrapackage imports
import pyx12.segment
import pyx12.x12file


def convert(filename, fd_out):
    """
    Convert a XML file in simple X12 form to an X12 file
    @param filename:  libxml2 requires a file name.  '-' gives stdin
    @type filename: string
    @param fd_out: Output file
    @type fd_out: file descripter
    """
    logger = logging.getLogger('pyx12')
    wr = pyx12.x12file.X12Writer(fd_out, '~', '*', ':', '\n', '^')
    doc = et.parse(filename)
    for node in doc.iter():
        if node.tag == 'seg':
            wr.Write(get_segment(node))
    return True


def get_segment(cSegment):
    """
    Build an X12 segment from a XML node
    """
    seg_id = cSegment.get('id')
    #seg_id = cSeg.findtext('data_ele')
    seg_data = pyx12.segment.Segment(seg_id, '~', '*', ':')
    for node in cSegment.iter():
        if node.tag == 'ele':
            ele_id = node.get('id')
            if node.text != '':
                seg_data.set(ele_id, node.text)
        elif node.tag == 'comp':
            for subele in node.findall('subele'):
                subele_id = subele.get('id')
                if subele.text is not None and subele.text != '':
                    seg_data.set(subele_id, subele.text)
    return seg_data



######################################################################
# Copyright 
#   John Holland <john@zoner.org>
# All rights reserved.
#
# This software is licensed as described in the file LICENSE.txt, which
# you should have received as part of this distribution.
#
######################################################################

"""
Loop and segment counter
"""
from collections import OrderedDict
import pyx12.path
from decorators import dump_args


class NodeCounter(object):
    """
    X12 Loop and Segment Node Counter
    """
    def __init__(self, initialCounts=None):
        if initialCounts is None:
            initialCounts = {}
        self._dict = OrderedDict()
        # copy constructor
        for k, v in initialCounts.items():
            self._dict[NodeCounter.makeX12Path(k)] = v

    #@dump_args
    def reset_to_node(self, xpath):
        """
        Pop to node, deleting all child counts
        Keep count of xpath node
        """
        parent = NodeCounter.makeX12Path(xpath)
        child_keys = list([x for x in self._dict.keys() if parent.is_child_path(x.format())])
        for k in child_keys:
            del self._dict[k]

    #@dump_args
    def increment(self, xpath):
        """
        Increment path count
        """
        k = NodeCounter.makeX12Path(xpath)
        if k in self._dict:
            self._dict[k] += 1
        else:
            self._dict[k] = 1

    #@dump_args
    def setCount(self, xpath, ct):
        """
        Set path count
        """
        k = NodeCounter.makeX12Path(xpath)
        self._dict[k] = ct

    def get_count(self, xpath):
        """
        Get path count
        """
        k = NodeCounter.makeX12Path(xpath)
        if k not in self._dict:
            return 0
        return self._dict[k]

    def getState(self):
        return self._dict

    @staticmethod
    def makeX12Path(xpath):
        if isinstance(xpath, pyx12.path.X12Path):
            return xpath
        else:
            return pyx12.path.X12Path(xpath)



#!/usr/bin/env python
import sys
import os
import os.path

libpath = os.path.abspath(os.path.join(os.path.dirname(__file__), '../..'))
if os.path.isdir(libpath):
    sys.path.insert(0, libpath)

import pyx12.x12file
import pyx12.error_handler
import sys


def main():
    for src_file in sys.argv[1:]:
        try:
            sys.stdout.write('Source filename: %s\n' % (src_file))
            src = pyx12.x12file.X12Reader(src_file)
            state = ''
            for c in src:
                if c.get_seg_id() == 'ISA':
                    sys.stdout.write('  ISA Sender: "%s"\t' %
                                     (c.get_value('ISA06')))
                    sys.stdout.write('ISA Receiver: "%s"\t' %
                                     (c.get_value('ISA08')))
                    if src.isa_usage == 'P':
                        sys.stdout.write(' PRODUCTION\t')
                    else:
                        sys.stdout.write(' TEST\t')
                    state = ''
                    sys.stdout.write('\n')
                elif c.get_seg_id() == 'GS':
                    sys.stdout.write('  GS Sender: "%s"\t' %
                                     (c.get_value('GS02')))
                    sys.stdout.write('GS Receiver: "%s"\t' %
                                     (c.get_value('GS03')))
                    sys.stdout.write('GS Type: "%s"\t' % (c.get_value('GS08')))
                    state = ''
                    sys.stdout.write('\n')
                elif c.get_seg_id() == 'ST':
                    sys.stdout.write('  ST ID: "%s"\t' % (c.get_value('ST02')))
                    sys.stdout.write('  ST Type: "%s"\t' %
                                     (c.get_value('ST01')))
                    state = ''
                    sys.stdout.write('\n')
                else:
                    state = ''
        except Exception:
            sys.stderr.write('File %s failed.' % (src_file))
            raise

        #self.gs_count = 0
        #self.st_count = 0
        #self.hl_count = 0
        #self.seg_count = 0
        #self.cur_line = 0
        #self.buffer = None


if __name__ == '__main__':
    sys.exit(not main())



######################################################################
# Copyright (c) Kalamazoo Community Mental Health Services,
#   John Holland <jholland@kazoocmh.org> <john@zoner.org>
# All rights reserved.
#
# This software is licensed as described in the file LICENSE.txt, which
# you should have received as part of this distribution.
#
######################################################################

#    $Id$

"""
Low level interface to an X12 data input stream.
Iterates over segment line strings.
Used by X12Reader.
"""

# Intrapackage imports
import pyx12.errors
import pyx12.segment

DEFAULT_BUFSIZE = 8 * 1024
ISA_LEN = 106


class RawX12File(object):
    """
    Interface to an X12 data file
    """

    def __init__(self, fin):
        """
        Initialize the file X12 file reader

        @param fin: an open, readable file object
        @type fin: open file object
        """
        self.fd = fin
        self.buffer = None
        line = self.fd.read(ISA_LEN)
        if line[:3] != 'ISA':
            err_str = "First line does not begin with 'ISA': %s" % line[:3]
            raise pyx12.errors.X12Error(err_str)
        if len(line) != ISA_LEN:
            err_str = 'ISA line is only %i characters' % len(line)
            raise pyx12.errors.X12Error(err_str)
        self.icvn = line[84:89]
        if self.icvn not in ('00401', '00501'):
            err_str = 'ISA Interchange Control Version Number is unknown: %s for %s' % (self.icvn, line)
            raise pyx12.errors.X12Error(err_str)
        self.seg_term = line[-1]
        self.ele_term = line[3]
        self.subele_term = line[-2]
        self.repetition_term = line[82] if self.icvn == '00501' else None
        self.buffer = line
        self.buffer += self.fd.read(DEFAULT_BUFSIZE)

    def __iter__(self):
        """
        Iterate over input lines
        Often, X12 files have a CR-LF after the segment delimiter.
        Split the input stream on the delimiter and remove any leading CR-LF
        """
        while True:
            if self.buffer.find(self.seg_term) == -1:
                # Need more data
                self.buffer += self.fd.read(DEFAULT_BUFSIZE)
            if self.buffer.find(self.seg_term) == -1:
                # Still have no segment terminator
                break
            # Get first segment in buffer
            (line, self.buffer) = self.buffer.split(self.seg_term, 1)
            line = line.lstrip('\n\r')
            if line == '':
                break
            yield(line)

    def get_term(self):
        """
        Get the original terminators

        @rtype: tuple(string, string, string, string)
        """
        return (self.seg_term, self.ele_term, self.subele_term, '\n', self.repetition_term)



######################################################################
# Copyright (c)
#   John Holland <john@zoner.org>
# All rights reserved.
#
# This software is licensed as described in the file LICENSE.txt, which
# you should have received as part of this distribution.
#
######################################################################

"""
Capture X12 Errors
"""

import logging
#from types import *
#import pdb
import tempfile
#import lxml
import os

# Intrapackage imports
from errors import EngineError
from xmlwriter import XMLWriter

#class error_node:
#    def __init__(self)
logger = logging.getLogger('pyx12.errh_xml')
#logger.setLevel(logging.DEBUG)
#logger.setLevel(logging.ERROR)

class err_handler(object):
    """
    The interface to the error handling structures.
    """
    def __init__(self, xml_out=None, basedir=None):
        """
        @param xml_out: Output filename, if None, will dump to tempfile
        @param basedir: working directory, where file will be created
        """
        if xml_out:
            self.filename = xml_out
            fd = open(xml_out, 'w')
        else:
            try:
                (fdesc, self.filename) = tempfile.mkstemp('.xml', 'pyx12_')
                fd = os.fdopen(fdesc, 'w+b')
                #fd = tempfile.NamedTemporaryFile()
                #self.filename = fd.name
            except:
                #self.filename = '997.tmp.xml'
                (fdesc, self.filename) = tempfile.mkstemp(suffix='.xml',
                                                          prefix='pyx12_', dir=basedir)
                fd = os.fdopen(fdesc, 'w+b')
                #fd = file(os.path.join(basedir, self.filename), 'w')
        self.cur_line = None
        self.errors = []
        if not fd:
            raise EngineError('Could not open temp error xml file')
        self.writer = XMLWriter(fd)
        #self.writer.doctype(
        #    u"x12simple", u"-//J Holland//DTD XML X12 Document
        #    Conversion1.0//EN//XML",
        #    u"%s" % (dtd_urn))
        self.writer.push("x12err")

    def __del__(self):
        while len(self.writer) > 0:
            self.writer.pop()

    def getFilename(self):
        return self.filename

    def handleErrors(self, err_list):
        """
        @param err_list: list of errors to apply
        """
        self.errors.extend(err_list)
        #for (err_type, err_cde, err_str, err_val, src_line) in err_list:
        #    if err_type == 'isa':
        #        self.isa_error(err_cde, err_str)
        #    elif err_type == 'gs':
        #        self.gs_error(err_cde, err_str)
        #    elif err_type == 'st':
        #        self.st_error(err_cde, err_str)
        #    elif err_type == 'seg':
        #        self.seg_error(err_cde, err_str, err_val, src_line)

    def getCurLine(self):
        """
        @return: Current file line number
        @rtype: int
        """
        return self.cur_line

    def Write(self, cur_line):
        """
        Generate XML for the segment data and matching map node

        """
        if len(self.errors) > 0:
            self.writer.push("seg", attrs={'line': '%i' % (cur_line)})
            for (err_type, err_cde, err_str, err_val, src_line) in self.errors:
                self.writer.push("err", attrs={"code": err_cde})
                #self.writer.elem(u"type", err_type)
                #self.writer.elem(u"code", err_cde)
                self.writer.elem("desc", err_str)
                if err_val:
                    self.writer.elem("errval", err_val)
                #self.writer.push(u"seg", {u'line': '%i'%(cur_line)})
                        #self.writer.elem(u'ele', seg_data.get_value('%02i' %
                        #(i+1)),
                        #    attrs={u'id': child_node.id})
                self.writer.pop()  # end err
            self.writer.pop()  # end segment
            self.errors = []


class ErrorErrhNull(Exception):
    """Class for errh_null errors."""


class errh_list(object):
    """
    A null error object - used for testing.
    Stores the current error in simple variables.
    """
    def __init__(self):
        #self.id = 'ROOT'
        self.errors = []
        #self.cur_node = self
        self.cur_line = 0
        #self.err_cde = None
        #self.err_str = None

    def get_errors(self):
        return self.errors

    def reset(self):
        self.errors = []

    def get_cur_line(self):
        """
        @return: Current file line number
        @rtype: int
        """
        return self.cur_line

    def set_cur_line(self, cur_line):
        """
        """
        self.cur_line = cur_line

#    def get_id(self):
#        """
#        @return: Error node type
#        @rtype: string
#        """
#        return self.id

    def add_isa_loop(self, seg, src):
        """
        """
        #raise ErrorErrhNull, 'add_isa loop'
        pass

    def add_gs_loop(self, seg, src):
        """
        """
        pass

    def add_st_loop(self, seg, src):
        """
        """
        pass

    def add_seg(self, map_node, seg, seg_count, cur_line, ls_id):
        """
        """
        pass

    def add_ele(self, map_node):
        """
        """
        pass

    def isa_error(self, err_cde, err_str):
        """
        @param err_cde: ISA level error code
        @type err_cde: string
        @param err_str: Description of the error
        @type err_str: string
        """
        self.errors.append(('isa', err_cde, err_str, None, None))
        sout = ''
        sout += 'Line:%i ' % (self.cur_line)
        sout += 'ISA:%s - %s' % (err_cde, err_str)
        logger.error(sout)

    def gs_error(self, err_cde, err_str):
        """
        @param err_cde: GS level error code
        @type err_cde: string
        @param err_str: Description of the error
        @type err_str: string
        """
        self.errors.append(('gs', err_cde, err_str, None, None))
        sout = ''
        sout += 'Line:%i ' % (self.cur_line)
        sout += 'GS:%s - %s' % (err_cde, err_str)
        logger.error(sout)

    def st_error(self, err_cde, err_str):
        """
        @param err_cde: Segment level error code
        @type err_cde: string
        @param err_str: Description of the error
        @type err_str: string
        """
        self.errors.append(('st', err_cde, err_str, None, None))
        sout = ''
        sout += 'Line:%i ' % (self.cur_line)
        sout += 'ST:%s - %s' % (err_cde, err_str)
        logger.error(sout)

    def seg_error(self, err_cde, err_str, err_value=None, src_line=None):
        """
        @param err_cde: Segment level error code
        @type err_cde: string
        @param err_str: Description of the error
        @type err_str: string
        """
        self.errors.append(('seg', err_cde, err_str, err_value, src_line))
        sout = ''
        sout += 'Line:%i ' % (self.cur_line)
        sout += 'SEG:%s - %s' % (err_cde, err_str)
        if err_value:
            sout += ' (%s)' % err_value
        logger.error(sout)

    def ele_error(self, err_cde, err_str, bad_value):
        """
        @param err_cde: Element level error code
        @type err_cde: string
        @param err_str: Description of the error
        @type err_str: string
        """
        self.errors.append(('ele', err_cde, err_str, bad_value, None))
        sout = ''
        sout += 'Line:%i ' % (self.cur_line)
        sout += 'ELE:%s - %s' % (err_cde, err_str)
        if bad_value:
            sout += ' (%s)' % (bad_value)
        logger.error(sout)

    def close_isa_loop(self, node, seg, src):
        """
        """
        pass

    def close_gs_loop(self, node, seg, src):
        """
        """
        pass

    def close_st_loop(self, node, seg, src):
        """
        """
        pass

    def find_node(self, atype):
        """
        Find the last node of a type
        """
        pass

    def get_parent(self):
        return None

#    def get_first_child(self):
#        """
#        """
#        if len(self.children) > 0:
#            return self.children[0]
#        else:
#            return None

    def get_next_sibling(self):
        """
        """
        return None

    def get_error_count(self):
        """
        """
        return len(self.errors)

    def is_closed(self):
        """
        @rtype: boolean
        """
        return True



######################################################################
# Copyright 
#   John Holland <john@zoner.org>
# All rights reserved.
#
# This software is licensed as described in the file LICENSE.txt, which
# you should have received as part of this distribution.
#
######################################################################

"""
"""

from errors import EngineError

isa_errors = ('000', '001', '002', '003', '004', '005', '006', '007', '008',
              '009', '010', '011', '012', '013', '014', '015', '016',
              '017', '018', '019', '020', '021', '022', '023', '024',
              '025', '026', '027', '028', '029', '030', '031')
seg_errors = ('1', '2', '3', '4', '5', '6', '7', '8')
ele_errors = ('1', '2', '3', '4', '5', '6', '7', '8', '9', '10')


class ErrorItem(object):
    """
    Wrap an X12 validation error
    """
    def __init__(self, err_type, err_cde, err_str):
        """
        @param err_type: At what level did the error occur
        @type err_type: string
        @param err_cde: Segment level error code
        @type err_cde: string
        @param err_str: Description of the error
        @type err_str: string
        """
        self.err_cde = err_cde
        self.err_str = err_str

    def getErrCde(self):
        return self.err_cde

    def getErrStr(self):
        return self.err_str


class ISAError(ErrorItem):
    def __init__(self, err_cde, err_str):
        ErrorItem.__init__(self, 'isa', err_cde, err_str)
        if self.err_cde not in isa_errors:
            raise EngineError('Invalid ISA level error code "%s"' %
                              (self.err_cde))


class SegError(ErrorItem):
    def __init__(self, err_cde, err_str, err_val=None):
        ErrorItem.__init__(self, 'seg', err_cde, err_str)
        self.err_val = err_val
        if self.err_cde not in seg_errors:
            raise EngineError('Invalid segment level error code "%s"' %
                              (self.err_cde))

    def getErrVal(self):
        return self.err_val


class EleError(ErrorItem):
    def __init__(self, err_cde, err_str, ele_idx, subele_idx=None,
                 err_val=None):
        ErrorItem.__init__(self, 'ele', err_cde, err_str)
        self.err_val = err_val
        self.ele_idx = ele_idx
        self.subele_idx = subele_idx
        if self.err_cde not in ele_errors:
            raise EngineError('Invalid element level error code "%s"' %
                              (self.err_cde))

    def getErrVal(self):
        return self.err_val

    def getEleIdx(self):
        return self.ele_idx

    def getSubeleIdx(self):
        return self.subele_idx



######################################################################
# Copyright (c) 2001-2011 
#   John Holland <john@zoner.org>
# All rights reserved.
#
# This software is licensed as described in the file LICENSE.txt, which
# you should have received as part of this distribution.
#
######################################################################

#    $Id: map_if.py 1472 2011-10-13 19:20:13Z johnholland $

"""
X12 syntax validation functions
"""


def is_syntax_valid(seg_data, syn):
    """
    Verifies the segment against the syntax
    @param seg_data: data segment instance
    @type seg_data: L{segment<segment.Segment>}
    @param syn: list containing the syntax type, and the indices of elements
    @type syn: list[string]
    @rtype: tuple(boolean, error string)
    """
    # handle intra-segment dependancies
    if len(syn) < 3:
        err_str = 'Syntax string must have at least two comparators (%s)' \
            % (syntax_str(syn))
        return (False, err_str)

    syn_code = syn[0]
    syn_idx = [int(s) for s in syn[1:]]

    if syn_code == 'P':
        count = 0
        for s in syn_idx:
            if len(seg_data) >= s and seg_data.get_value('%02i' % (s)) != '':
                count += 1
        if count != 0 and count != len(syn_idx):
            err_str = 'Syntax Error (%s): If any of %s is present, then all are required'\
                % (syntax_str(syn), syntax_ele_id_str(seg_data.get_seg_id(), syn_idx))
            return (False, err_str)
        else:
            return (True, None)
    elif syn_code == 'R':
        count = 0
        for s in syn_idx:
            if len(seg_data) >= s and seg_data.get_value('%02i' % (s)) != '':
                count += 1
        if count == 0:
            err_str = 'Syntax Error (%s): At least one element is required' % \
                (syntax_str(syn))
            return (False, err_str)
        else:
            return (True, None)
    elif syn_code == 'E':
        count = 0
        for s in syn_idx:
            if len(seg_data) >= s and seg_data.get_value('%02i' % (s)) != '':
                count += 1
        if count > 1:
            err_str = 'Syntax Error (%s): At most one of %s may be present'\
                % (syntax_str(syn), syntax_ele_id_str(seg_data.get_seg_id(), syn_idx))
            return (False, err_str)
        else:
            return (True, None)
    elif syn_code == 'C':
        # If the first is present, then all others are required
        if len(seg_data) >= syn_idx[0] and seg_data.get_value('%02i' % (syn_idx[0])) != '':
            count = 0
            for s in syn_idx[1:]:
                if len(seg_data) >= s and seg_data.get_value('%02i' % (s)) != '':
                    count += 1
            if count != len(syn_idx) - 1:
                if len(syn_idx[1:]) > 1: verb = 'are'
                else:
                    verb = 'is'
                err_str = 'Syntax Error (%s): If %s%02i is present, then %s %s required'\
                    % (syntax_str(syn), seg_data.get_seg_id(), syn_idx[0],
                       syntax_ele_id_str(seg_data.get_seg_id(), syn_idx[1:]), verb)
                return (False, err_str)
            else:
                return (True, None)
        else:
            return (True, None)
    elif syn_code == 'L':
        if len(seg_data) > syn_idx[0] - 1 and seg_data.get_value('%02i' % (syn_idx[0])) != '':
            count = 0
            for s in syn_idx[1:]:
                if len(seg_data) >= s and seg_data.get_value('%02i' % (s)) != '':
                    count += 1
            if count == 0:
                err_str = 'Syntax Error (%s): If %s%02i is present, then at least one of '\
                    % (syntax_str(syn), seg_data.get_seg_id(), syn_idx[0])
                err_str += syntax_ele_id_str(
                    seg_data.get_seg_id(), syn_idx[1:])
                err_str += ' is required'
                return (False, err_str)
            else:
                return (True, None)
        else:
            return (True, None)
    #raise EngineError
    return (False, 'Syntax Type %s Not Found' % (syntax_str(syn)))


def syntax_str(syntax):
    """
    @rtype: string
    """
    output = syntax[0]
    for i in syntax[1:]:
        output += '%02i' % (i)
    return output


def syntax_ele_id_str(seg_id, ele_pos_list):
    """
    @rtype: string
    """
    output = ''
    output += '%s%02i' % (seg_id, ele_pos_list[0])
    for i in range(len(ele_pos_list) - 1):
        if i == len(ele_pos_list) - 2:
            output += ' or %s%02i' % (seg_id, ele_pos_list[i + 1])
        else:
            output += ', %s%02i' % (seg_id, ele_pos_list[i + 1])
    return output



######################################################################
# Copyright 
#   John Holland <john@zoner.org>
# All rights reserved.
#
# This software is licensed as described in the file LICENSE.txt, which
# you should have received as part of this distribution.
#
######################################################################

"""
Create a XML rendering of the X12 document
"""

from os.path import commonprefix
import logging

# Intrapackage imports
from errors import EngineError
from x12xml import x12xml
from map_walker import pop_to_parent_loop

logger = logging.getLogger('pyx12.x12xml.simple')


class x12xml_simple(x12xml):
    def __init__(self, fd, dtd_urn=None):
        x12xml.__init__(self, fd, "x12simple", dtd_urn)
        self.last_path = []

    def __del__(self):
        while len(self.writer) > 0:
            self.writer.pop()

    def seg(self, seg_node, seg_data):
        """
        Generate XML for the segment data and matching map node

        @param seg_node: Map Node
        @type seg_node: L{node<map_if.x12_node>}
        @param seg_data: Segment object
        @type seg_data: L{segment<segment.Segment>}
        """
        if not seg_node.is_segment():
            raise EngineError('Node must be a segment')
        parent = pop_to_parent_loop(seg_node)  # Get enclosing loop
        # check path for new loops to be added
        cur_path = self._path_list(parent.get_path())
        #if seg_node.id == 'GS':
        #    import ipdb; ipdb.set_trace()
        if self.last_path == cur_path and seg_node.is_first_seg_in_loop():
            # loop repeat
            self.writer.pop()
            (xname, attrib) = self._get_loop_info(cur_path[-1])
            self.writer.push(xname, attrib)
        else:
            last_path = self.last_path
            match_idx = self._get_path_match_idx(last_path, cur_path)
            root_path = self._path_list(commonprefix(['/'.join(cur_path), '/'.join(last_path)]))
            if seg_node.is_first_seg_in_loop() and root_path == cur_path:
                match_idx -= 1
            for i in range(len(last_path) - 1, match_idx - 1, -1):
                self.writer.pop()
            for i in range(match_idx, len(cur_path)):
                (xname, attrib) = self._get_loop_info(cur_path[i])
                self.writer.push(xname, attrib)
        seg_node_id = self._get_node_id(seg_node, parent, seg_data)
        (xname, attrib) = self._get_seg_info(seg_node_id)
        self.writer.push(xname, attrib)
        for i in range(len(seg_data)):
            child_node = seg_node.get_child_node_by_idx(i)
            if child_node.usage == 'N' or seg_data.get('%02i' % (i + 1)).is_empty():
                pass  # Do not try to ouput for invalid or empty elements
            elif child_node.is_composite():
                (xname, attrib) = self._get_comp_info(seg_node_id)
                self.writer.push(xname, attrib)
                comp_data = seg_data.get('%02i' % (i + 1))
                for j in range(len(comp_data)):
                    subele_node = child_node.get_child_node_by_idx(j)
                    (xname, attrib) = self._get_subele_info(subele_node.id)
                    self.writer.elem(xname, comp_data[j].get_value(), attrib)
                self.writer.pop()  # end composite
            elif child_node.is_element():
                if seg_data.get_value('%02i' % (i + 1)) == '':
                    pass
                    #self.writer.empty(u"ele", attrs={u'id': child_node.id})
                else:
                    (xname, attrib) = self._get_ele_info(child_node.id)
                    self.writer.elem(xname, seg_data.get_value('%02i' % (i + 1)), attrib)
            else:
                raise EngineError('Node must be a either an element or a composite')
        self.writer.pop()  # end segment
        self.last_path = cur_path

    def _get_loop_info(self, loop_id):
        """
        Override loop node value
        """
        return ("loop", {'id': loop_id})

    def _get_seg_info(self, seg_id):
        """
        Override segment node value
        """
        return ("seg", {'id': seg_id})

    def _get_comp_info(self, comp_id):
        """
        Override composite node value
        """
        return ("comp",  {'id': comp_id})

    def _get_ele_info(self, ele_id):
        """
        Override element node value
        """
        return ("ele", {'id': ele_id})

    def _get_subele_info(self, subele_id):
        """
        Override sub-element node value
        """
        return ("subele", {'id': subele_id})



######################################################################
# Copyright (c) 2001-2011 
#   John Holland <john@zoner.org>
# All rights reserved.
#
# This software is licensed as described in the file LICENSE.txt, which
# you should have received as part of this distribution.
#
######################################################################

"""
Interface to normalized Data Elements
"""

import os.path
import logging
import xml.etree.cElementTree as et
from pkg_resources import resource_stream

# Intrapackage imports
from pyx12.errors import EngineError


class DataElementsError(Exception):
    """Class for data elements module errors."""


class DataElements(object):
    """
    Interface to normalized Data Elements
    """

    def __init__(self, base_path=None):
        """
        Initialize the list of data elements
        @param base_path: Override directory containing dataele.xml.  If None,
            uses package resource folder
        @type base_path: string

        @note: self.dataele - map to the data element
        {ele_num: {data_type, min_len, max_len, name}}
        """
        logger = logging.getLogger('pyx12')
        self.dataele = {}
        dataele_file = 'dataele.xml'
        if base_path is not None:
            logger.debug("Looking for data element definition file '{}' in map_path '{}'".format(dataele_file, base_path))
            fd = open(os.path.join(base_path, dataele_file))
        else:
            logger.debug("Looking for data element definition file '{}' in pkg_resources".format(dataele_file))
            fd = resource_stream(__name__, os.path.join('map', dataele_file))
        for eElem in et.parse(fd).iter('data_ele'):
            ele_num = eElem.get('ele_num')
            data_type = eElem.get('data_type')
            min_len = int(eElem.get('min_len'))
            max_len = int(eElem.get('max_len'))
            name = eElem.get('name')
            self.dataele[ele_num] = {'data_type': data_type, 'min_len':
                                     min_len, 'max_len': max_len, 'name': name}

    def get_by_elem_num(self, ele_num):
        """
        Get the element characteristics for the indexed element code
        @param ele_num: the data element code
        @type ele_num: string
        @return: {data_type, min_len, max_len, name}
        @rtype: dict
        """
        if not ele_num:
            raise EngineError('Bad data element %s' % (ele_num))
        if ele_num not in self.dataele:
            raise EngineError('Data Element "%s" is not defined' % (ele_num))
        return self.dataele[ele_num]

    def __repr__(self):
        for ele_num in list(self.dataele.keys()):
            print((self.dataele[ele_num]))

    def debug_print(self):
        """
        Debug print data elements
        """
        self.__repr__()



#!/usr/bin/env python

######################################################################
# Copyright 
#   John Holland <john@zoner.org>
# All rights reserved.
#
# This software is licensed as described in the file LICENSE.txt, which
# you should have received as part of this distribution.
#
######################################################################

"""
Create an X12 document from a XML data file
"""

import os
import os.path
import sys
import logging
import codecs

# Intrapackage imports
libpath = os.path.abspath(os.path.join(os.path.dirname(__file__), '../..'))
if os.path.isdir(libpath):
    sys.path.insert(0, libpath)
import pyx12.segment
import pyx12.xmlx12_simple

#Global Variables
__author__ = pyx12.__author__
__status__ = pyx12.__status__
__version__ = pyx12.__version__
__date__ = pyx12.__date__


def main():
    """Script main program."""
    import argparse
    parser = argparse.ArgumentParser(description='XML to X12 conversion')
    parser.add_argument('--log-file', '-l', action='store', dest="logfile", default=None)
    parser.add_argument('--verbose', '-v', action='count')
    parser.add_argument('--debug', '-d', action='store_true')
    parser.add_argument('--quiet', '-q', action='store_true')
    parser.add_argument('--outputfile', '-o', action='store', help="X12 target filename")
    parser.add_argument('--version', action='version', version='{prog} {version}'.format(prog=parser.prog, version=__version__))
    parser.add_argument('input_file')
    args = parser.parse_args()

    logger = logging.getLogger('pyx12')
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')

    stdout_hdlr = logging.StreamHandler()
    stdout_hdlr.setFormatter(formatter)
    logger.addHandler(stdout_hdlr)
    logger.setLevel(logging.INFO)

    if args.debug:
        logger.setLevel(logging.DEBUG)
    if args.verbose > 0:
        logger.setLevel(logging.DEBUG)
    if args.quiet:
        logger.setLevel(logging.ERROR)
    if args.logfile:
        try:
            hdlr = logging.FileHandler(args.logfile)
            hdlr.setFormatter(formatter)
            logger.addHandler(hdlr)
        except IOError:
            logger.exception('Could not open log file: %s' % (args.logfile))

    if args.input_file:
        try:
            fd_source = open(args.input_file)
        except:
            logger.error('Could not open file %s' % (args.input_file))
            return False
    else:
        fd_source = sys.stdin

    if args.outputfile:
        try:
            fd_x12 = codecs.open(args.outputfile, mode='w', encoding='ascii')
        except:
            logger.error('Could not open file %s' % (args.outputfile))
            return False
    else:
        fd_x12 = sys.stdout

    try:
        result = pyx12.xmlx12_simple.convert(fd_source, fd_x12)
        if not result:
            logger.error('Input file had errors.')
            return False
    except KeyboardInterrupt:
        print "\n[interrupt]"

    return True

if __name__ == '__main__':
    sys.exit(not main())



#!/usr/bin/env python

"""
Format a X12 document.  If the option -e is used, it adds newlines.
If no source file is given, read from stdin.
If no ouput filename is given with -o,  write to stdout.
"""

import sys
import os.path
import codecs
import tempfile
import logging

# Intrapackage imports
libpath = os.path.abspath(os.path.join(os.path.dirname(__file__), '../..'))
if os.path.isdir(libpath):
    sys.path.insert(0, libpath)
import pyx12
import pyx12.x12file

__author__ = pyx12.__author__
__status__ = pyx12.__status__
__version__ = pyx12.__version__
__date__ = pyx12.__date__


def main():
    import argparse
    parser = argparse.ArgumentParser(description='X12 Validation')
    parser.add_argument('--verbose', '-v', action='count')
    parser.add_argument('--quiet', '-q', action='store_true')
    parser.add_argument('--debug', '-d', action='store_true')
    parser.add_argument('--eol', '-e', action='store_true', help="Add eol to each segment line")
    parser.add_argument('--inplace', '-i', action='store_true', help="Make changes to files in place")
    parser.add_argument('--fixcounting', '-f', action='store_true', help="Try to fix counting errors")
    #parser.add_argument('--fixwhitespace', '-w', action='store_true', help="Try to fix extra whitespace errors.")
    parser.add_argument('--output', '-o', action='store', dest="outputfile", default=None, help="Output filename.  Defaults to stdout")
    parser.add_argument('--version', action='version', version='{prog} {version}'.format(prog=parser.prog, version=__version__))
    parser.add_argument('input_files', nargs='*')
    args = parser.parse_args()

    logger = logging.getLogger()
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    stdout_hdlr = logging.StreamHandler()
    stdout_hdlr.setFormatter(formatter)
    logger.addHandler(stdout_hdlr)
    logger.setLevel(logging.INFO)

    eol = '\n' if args.eol else ''
    for file_in in args.input_files:
        if not os.path.isfile(file_in):
            logger.error('Could not open file "%s"' % (file_in))

        fd_out = tempfile.TemporaryFile()
        src = pyx12.x12file.X12Reader(file_in)
        for seg_data in src:
            if args.fixcounting:
                err_codes = [(x[1]) for x in src.pop_errors()]
                if seg_data.get_seg_id() == 'IEA' and '021' in err_codes:
                    seg_data.set('IEA01', '%i' % (src.gs_count))
                elif seg_data.get_seg_id() == 'GE' and '5' in err_codes:
                    seg_data.set('GE01', '%i' % (src.st_count))
                elif seg_data.get_seg_id() == 'SE' and '4' in err_codes:
                    seg_data.set('SE01', '%i' % (src.seg_count + 1))
                elif seg_data.get_seg_id() == 'HL' and 'HL1' in err_codes:
                    seg_data.set('HL01', '%i' % (src.hl_count))
            #if args.fixwhitespace:
            #    err_codes = [(x[1]) for x in src.pop_errors()]
            #    if 'SEG1' in err_codes:
            fd_out.write(seg_data.format() + eol)
        if eol == '':
            fd_out.write('\n')

        fd_out.seek(0)
        if args.outputfile:
            fd_out = codecs.open(args.outputfile, mode='w', encoding='ascii')
        else:
            if args.inplace:
                with codecs.open(file_in, mode='w', encoding='ascii') as fd_orig:
                    fd_orig.write(fd_out.read())
            else:
                sys.stdout.write(fd_out.read())
    return True

if __name__ == '__main__':
    sys.exit(not main())



######################################################################
# Copyright 
#   John Holland <john@zoner.org>
# All rights reserved.
#
# This software is licensed as described in the file LICENSE.txt, which
# you should have received as part of this distribution.
#
######################################################################

"""
Interface to X12 Errors
"""

import logging

# Intrapackage imports
from errors import IterOutOfBounds  # , IterDone

logger = logging.getLogger('pyx12.error_handler')


class err_iter(object):
    """
    Iterate over the error tree

    Implements an odd iterator???
    """

    def __init__(self, errh):
        """
        @param errh: Error_handler instance
        @type errh: L{error_handler.err_handler}
        """
        self.errh = errh
        self.cur_node = errh
        self.visit_stack = []

    def first(self):
        self.cur_node = self.errh

    def next(self):
        self.__next__()

    def __next__(self):
        #If at previosly visited branch, do not do children
        if self.cur_node in self.visit_stack:
            node = None
        else:
            node = self.cur_node.get_first_child()
        if node is not None:
            self.visit_stack.append(self.cur_node)
            self.cur_node = node
        else:
            node = self.cur_node.get_next_sibling()
            if node is not None:
                self.cur_node = node
            else:
                if not self.cur_node.is_closed():
                    raise IterOutOfBounds
                node = self.cur_node.get_parent()
                if node is None:
                    raise IterOutOfBounds
                if not node.is_closed():
                    raise IterOutOfBounds
                if self.cur_node in self.visit_stack:
                    del self.visit_stack[-1]
                self.cur_node = node
                if node.id == 'ROOT':
                    raise IterOutOfBounds
                #    raise IterDone

    def get_cur_node(self):
        return self.cur_node


class err_handler(object):
    """
    The interface to the error handling structures.
    """
    def __init__(self):
        """
        """

        self.id = 'ROOT'
        #self.isa_loop_count = 0
        self.children = []
        self.cur_node = self
        self.cur_isa_node = None
        self.cur_gs_node = None
        self.cur_st_node = None
        self.cur_seg_node = None
        self.seg_node_added = False
        self.cur_ele_node = None
        self.cur_line = 0

    def accept(self, visitor):
        """
        Params:     visitor - ref to visitor class
        """
        visitor.visit_root_pre(self)
        for child in self.children:
            child.accept(visitor)
        visitor.visit_root_post(self)

    def handle_errors(self, err_list):
        """
        @param err_list: list of errors to apply
        """
        for (err_type, err_cde, err_str, err_val, src_line) in err_list:
            if err_type == 'isa':
                self.isa_error(err_cde, err_str)
            elif err_type == 'gs':
                self.gs_error(err_cde, err_str)
            elif err_type == 'st':
                self.st_error(err_cde, err_str)
            elif err_type == 'seg':
                self.seg_error(err_cde, err_str, err_val, src_line)

    def get_cur_line(self):
        """
        @return: Current file line number
        @rtype: int
        """
        return self.cur_line

    def get_id(self):
        """
        """
        return self.id

    def add_isa_loop(self, seg_data, src):
        """
        @param seg_data: Segment object
        @type seg_data: L{segment<segment.Segment>}
        """
        #logger.debug('add_isa loop')
        self.children.append(err_isa(self, seg_data, src))
        self.cur_isa_node = self.children[-1]
        self.cur_seg_node = self.cur_isa_node
        self.seg_node_added = True

    def add_gs_loop(self, seg_data, src):
        """
        @param seg_data: Segment object
        @type seg_data: L{segment<segment.Segment>}
        """
        #logger.debug('add_gs loop')
        parent = self.cur_isa_node
        parent.children.append(err_gs(parent, seg_data, src))
        self.cur_gs_node = parent.children[-1]
        self.cur_seg_node = self.cur_gs_node
        self.seg_node_added = True

    def add_st_loop(self, seg_data, src):
        """
        @param seg_data: Segment object
        @type seg_data: L{segment<segment.Segment>}
        """
        #logger.debug('add_st loop')
        parent = self.cur_gs_node
        parent.children.append(err_st(parent, seg_data, src))
        self.cur_st_node = parent.children[-1]
        self.cur_seg_node = self.cur_st_node
        self.seg_node_added = True

    def add_seg(self, map_node, seg_data, seg_count, cur_line, ls_id):
        """
        @param map_node: current segment node
        @type map_node: L{node<map_if.segment_if>}
        @param seg_data: Segment object
        @type seg_data: L{segment<segment.Segment>}
        @param seg_count: Count of current segment in the ST Loop
        @type seg_count: int
        @param cur_line: Current line number in the file
        @type cur_line: int
        @param ls_id: The current LS loop identifier
        @type ls_id: string
        """
        parent = self.cur_st_node
        self.cur_seg_node = err_seg(
            parent, map_node, seg_data, seg_count, cur_line, ls_id)
        self.seg_node_added = False
        #logger.debug('add_seg: %s' % map_node.name)
        #if len(parent.children) > 0:
        #    if parent.children[-1].err_count() == 0:
        #        del parent.children[-1]
        #        logger.debug('del seg_data: %s' % map_node.name)
        #parent.children.append(err_seg(parent, map_node, seg_data, src))

    def _add_cur_seg(self):
        """
        """
        #pdb.set_trace()
        if not self.seg_node_added:
            self.cur_st_node.children.append(self.cur_seg_node)
            self.seg_node_added = True

    def add_ele(self, map_node):
        """
        """
        if self.cur_seg_node.id == 'ISA':
            self.cur_ele_node = err_ele(self.cur_isa_node, map_node)
        elif self.cur_seg_node.id == 'GS':
            self.cur_ele_node = err_ele(self.cur_gs_node, map_node)
        elif self.cur_seg_node.id == 'ST':
            self.cur_ele_node = err_ele(self.cur_st_node, map_node)
        else:
            self.cur_ele_node = err_ele(self.cur_seg_node, map_node)
        self.ele_node_added = False

    def _add_cur_ele(self):
        """
        """
        self._add_cur_seg()
        if not self.ele_node_added and self.cur_seg_node is not None:
            self.cur_seg_node.elements.append(self.cur_ele_node)
            self.ele_node_added = True
        #logger.debug('----  add_ele: %s' % self.cur_seg_node.elements[-1].name)

    def isa_error(self, err_cde, err_str):
        """
        @param err_cde: ISA level error code
        @type err_cde: string
        @param err_str: Description of the error
        @type err_str: string
        """
        sout = ''
        sout += 'Line:%i ' % (self.cur_isa_node.get_cur_line())
        sout += 'ISA:%s - %s' % (err_cde, err_str)
        logger.error(sout)
        self.cur_isa_node.add_error(err_cde, err_str)

    def gs_error(self, err_cde, err_str):
        """
        @param err_cde: GS level error code
        @type err_cde: string
        @param err_str: Description of the error
        @type err_str: string
        """
        sout = ''
        sout += 'Line:%i ' % (self.cur_gs_node.get_cur_line())
        sout += 'GS:%s - %s' % (err_cde, err_str)
        logger.error(sout)
        self.cur_gs_node.add_error(err_cde, err_str)

    def st_error(self, err_cde, err_str):
        """
        @param err_cde: Segment level error code
        @type err_cde: string
        @param err_str: Description of the error
        @type err_str: string
        """
        sout = ''
        sout += 'Line:%i ' % (self.cur_st_node.get_cur_line())
        sout += 'ST:%s - %s' % (err_cde, err_str)
        logger.error(sout)
        self.cur_st_node.add_error(err_cde, err_str)

    def seg_error(self, err_cde, err_str, err_value=None, src_line=None):
        """
        @param err_cde: Segment level error code
        @type err_cde: string
        @param err_str: Description of the error
        @type err_str: string
        """
        sout = ''
        try:
            self._add_cur_seg()
            self.cur_seg_node.add_error(err_cde, err_str, err_value)
        except:
            sout += 'No current segment in error_handler. '
        if src_line:
            sout += 'Line:%i ' % (src_line)
        else:
            if self.cur_seg_node is not None:
                sout += 'Line:%i ' % (self.cur_seg_node.get_cur_line())
        sout += 'SEG:%s - %s' % (err_cde, err_str)
        if err_value:
            sout += ' (%s)' % err_value
        logger.error(sout)

    def ele_error(self, err_cde, err_str, bad_value, refdes=None):
        """
        @param err_cde: Element level error code
        @type err_cde: string
        @param err_str: Description of the error
        @type err_str: string
        """
        self._add_cur_ele()
        self.cur_ele_node.add_error(
            err_cde, err_str, bad_value)  # , pos, data_ele)
        sout = ''
        sout += 'Line:%i ' % (self.cur_seg_node.get_cur_line())
        sout += 'ELE:%s - %s' % (err_cde, err_str)
        if bad_value:
            sout += ' (%s)' % (bad_value)
        logger.error(sout)
        #print self.cur_ele_node.errors

    def close_isa_loop(self, node, seg, src):
        """
        """
        self.cur_isa_node.close(node, seg, src)
        self.cur_seg_node = self.cur_isa_node
        self.seg_node_added = True

    def close_gs_loop(self, node, seg, src):
        """
        """
        self.cur_gs_node.close(node, seg, src)
        self.cur_seg_node = self.cur_gs_node
        self.seg_node_added = True

    def close_st_loop(self, node, seg, src):
        """
        """
        self.cur_st_node.close(node, seg, src)
        self.cur_seg_node = self.cur_st_node
        self.seg_node_added = True

    def find_node(self, type):
        """
        Find the last node of a type
        """
        new_node = self.cur_node
        node_order = {'ROOT': 1, 'ISA': 2, 'GS': 3, 'ST': 4, 'SEG':
                      5, 'ELE': 6}
        while node_order[type] > new_node[new_node.get_id()]:
            new_node = new_node.get_parent()
        #walk error tree to find place to append
        #if type == 'ISA':
        #return node

#    def update_node(self, obj):
#        self.children[-1].update_node(obj)

    def _get_last_child(self):
        """
        """
        if len(self.children) != 0:
            return self.children[-1]
        else:
            return None

    def get_parent(self):
        return None

    def get_error_count(self):
        """
        """
        count = 0
        for child in self.children:
            count += child.get_error_count()
        return count

    def get_first_child(self):
        """
        """
        if len(self.children) > 0:
            return self.children[0]
        else:
            return None

    def get_next_sibling(self):
        """
        """
        return None
        #raise IterDone

    def __next__(self):
        """
        Return the next error node
        """
        for child in self.children:
            yield child

    def is_closed(self):
        """
        @rtype: boolean
        """
        return True

    def __repr__(self):
        """
        """
        return '%i: %s' % (-1, self.id)


class err_node(object):
    def __init__(self, parent):
        """
        """
        self.parent = parent
        self.id = None
        self.children = []
        self.cur_line = -1
        self.errors = []

    def accept(self, visitor):
        """
        """
        pass

#    def update_node(self, obj):
#        pass

    def get_cur_line(self):
        """
        @return: Current file line number
        @rtype: int
        """
        return self.cur_line

    def get_id(self):
        """
        """
        return self.id

    def get_parent(self):
        """
        """
        return self.parent

    def _get_last_child(self):
        """
        """
        if len(self.children) != 0:
            return self.children[-1]
        else:
            return None

    def get_next_sibling(self):
        """
        """
        #if self.id == 'ROOT': raise EngineError
        bFound = False
        for sibling in self.parent.children:
            if bFound:
                return sibling
            if sibling is self:
                bFound = True
        return None
        #raise IterOutOfBounds

    def get_first_child(self):
        """
        """
        if len(self.children) > 0:
            return self.children[0]
        else:
            return None

    def get_error_count(self):
        """
        """
        count = 0
        for child in self.children:
            count += child.get_error_count()
        return count

    def get_error_list(self, seg_id, pre=False):
        """
        """
        return self.errors

    def is_closed(self):
        """
        @rtype: boolean
        """
        return True


class err_isa(err_node):
    """
    Holds source ISA loop errors
    """

    def __init__(self, parent, seg_data, src):
        """
        @param seg_data: Segment object
        @type seg_data: L{segment<segment.Segment>}
        @param src: X12file source
        @type src: L{X12file<x12file.X12Reader>}
        """
        self.seg_data = seg_data
        self.isa_id = src.get_isa_id()
        self.cur_line_isa = src.get_cur_line()
        self.cur_line_iea = None

        self.isa_trn_set_id = seg_data.get_value('ISA13')
        self.ta1_req = seg_data.get_value('ISA14')
        self.orig_date = seg_data.get_value('ISA09')
        self.orig_time = seg_data.get_value('ISA10')
        self.id = 'ISA'

        self.parent = parent
        self.children = []
        self.errors = []
        self.elements = []

    def is_closed(self):
        """
        @rtype: boolean
        """
        if self.cur_line_iea:
            return True
        else:
            return False

    def accept(self, visitor):
        """
        Params:     visitor - ref to visitor class
        """
        visitor.visit_isa_pre(self)
        for child in self.children:
            child.accept(visitor)
        visitor.visit_isa_post(self)

    def add_error(self, err_cde, err_str):
        """
        @param err_cde: Error code
        @type err_cde: string
        @param err_str: Description of the error
        @type err_str: string
        """
        self.errors.append((err_cde, err_str))

    def close(self, node, seg, src):
        self.cur_line_iea = src.get_cur_line()

    def get_cur_line(self):
        """
        @return: Current file line number
        @rtype: int
        """
        if self.cur_line_iea:
            return self.cur_line_iea
        else:
            return self.cur_line_isa

    def get_error_count(self):
        """
        """
        count = 0
        for ele in self.elements:
            count += ele.get_error_count()
        for child in self.children:
            count += child.get_error_count()
        return count + len(self.errors)

    def get_error_list(self, seg_id, pre=False):
        """
        """
        if seg_id == 'ISA':
            return [err for err in self.errors if 'ISA' in err[0]]
        elif seg_id == 'IEA':
            return [err for err in self.errors if 'IEA' in err[0]]
        else:
            return []
        #err_list = []
        #for err in self.errors:
        #    if seg_id in err[0]:
        #        err_list.append(err)
        #return err_list

    def __next__(self):
        """
        Return the next error node
        """
        for child in self.children:
            yield child
        next(self.parent)

    def __repr__(self):
        return '%i: %s' % (self.get_cur_line(), self.id)


class err_gs(err_node):
    """
    Holds source GS loop information
    """

    def __init__(self, parent, seg_data, src):
        """
        @param seg_data: Segment object
        @type seg_data: L{segment<segment.Segment>}
        @param src: X12file source
        @type src: L{X12file<x12file.X12Reader>}

        """
        self.seg_data = seg_data
        self.isa_id = src.get_isa_id()
        self.cur_line_gs = src.get_cur_line()
        self.cur_line_ge = None
        self.gs_control_num = src.get_gs_id()
        self.fic = self.seg_data.get_value('GS01')
        self.vriic = self.seg_data.get_value('GS08')
        self.id = 'GS'

        self.st_loops = []

        # From GE loop
        self.ack_code = None  # AK901
        self.st_count_orig = 0  # AK902
        self.st_count_recv = 0  # AK903
        #self.st_count_accept = None # AK904

        self.parent = parent
        self.children = []
        self.errors = []
        self.elements = []

    def accept(self, visitor):
        """
        Params:     visitor - ref to visitor class
        """
        visitor.visit_gs_pre(self)
        for child in self.children:
            child.accept(visitor)
        visitor.visit_gs_post(self)

    def add_error(self, err_cde, err_str):
        """
        @param err_cde: Error code
        @type err_cde: string
        @param err_str: Description of the error
        @type err_str: string
        """
        self.errors.append((err_cde, err_str))

    def close(self, node, seg_data, src):
        """
        """
        # From GE loop
        self.cur_line_ge = src.get_cur_line()

        self.ack_code = self._get_ack_code()

        if seg_data is None:
            self.st_count_orig = 0
        else:
            self.st_count_orig = int(seg_data.get_value('GE01'))  # AK902
        self.st_count_recv = src.st_count  # AK903
        #self.st_count_accept = self.st_count_recv - len(self.children) # AK904

    def _get_ack_code(self):
        for child in self.children:
            if child.get_error_count() > 0:
                return 'R'
        #err_codes = map(lambda x:x[0], self.errors)
        #if '1' in err_codes: return 'R'
        #elif '2' in err_codes: return 'R'
        #elif '3' in err_codes: return 'R'
        #elif '4' in err_codes: return 'R'
        #elif '5' in err_codes: return 'E'
        #elif '6' in err_codes: return 'E'
        if len(self.errors) > 0:
            return 'R'
        return 'A'

    def count_failed_st(self):
        ct = 0
        for child in self.children:
            if child.ack_code not in ['A', 'E']:
                ct += 1
        return ct

    def get_cur_line(self):
        """
        @return: Current file line number
        @rtype: int
        """
        if self.cur_line_ge:
            return self.cur_line_ge
        else:
            return self.cur_line_gs

    def get_error_count(self):
        """
        """
        count = 0
        for ele in self.elements:
            count += ele.get_error_count()
        for child in self.children:
            count += child.get_error_count()
        return count + len(self.errors)

    def get_error_list(self, seg_id, pre=False):
        """
        """
        if seg_id == 'GS':
            return [err for err in self.errors if err[0] in ('6')]
        elif seg_id == 'GE':
            return [err for err in self.errors if err[0] not in ('6')]
        else:
            return []

    def is_closed(self):
        """
        @rtype: boolean
        """
        if self.cur_line_ge:
            return True
        else:
            return False

    def __next__(self):
        """
        Return the next error node
        """
        for child in self.children:
            yield child
        next(self.parent)

    def __repr__(self):
        return '%i: %s' % (self.get_cur_line(), self.id)


class err_st(err_node):
    """
    ST loops

    Needs:
        1. Transaction set id code (837, 834)
        2. Transaction set control number
        3. trn set error codes
        4. At SE, Determine final ack code
    """

    def __init__(self, parent, seg_data, src):
        """
        @param seg_data: Segment object
        @type seg_data: L{segment<segment.Segment>}
        @param src: X12file source
        @type src: L{X12file<x12file.X12Reader>}
        """
        self.seg_data = seg_data
        self.trn_set_control_num = src.get_st_id()
        self.cur_line_st = src.get_cur_line()
        self.cur_line_se = None
        self.trn_set_id = seg_data.get_value('ST01')
        self.vriic = seg_data.get_value('ST03')
        self.id = 'ST'

        self.ack_code = 'R'
        self.parent = parent
        self.children = []
        self.errors = []
        self.elements = []
        #self.rejected = None

    def accept(self, visitor):
        """
        Params:     visitor - ref to visitor class
        """
        visitor.visit_st_pre(self)
        for child in self.children:
            child.accept(visitor)
        visitor.visit_st_post(self)

    def add_error(self, err_cde, err_str):
        """
        @param err_cde: Error code
        @type err_cde: string
        @param err_str: Description of the error
        @type err_str: string
        """
        self.errors.append((err_cde, err_str))

    def close(self, node, seg_data, src):
        """
        Close ST loop

        @param node: SE node
        @type node: L{node<map_if.x12_node>}
        @param seg_data: Segment object
        @type seg_data: L{segment<segment.Segment>}
        @param src: X12file source
        @type src: L{X12file<x12file.X12Reader>}
        """
        self.cur_line_se = src.get_cur_line()
        #pdb.set_trace()
        if self.err_count() > 0:
            self.ack_code = 'R'
        else:
            self.ack_code = 'A'

    def err_count(self):
        """
        @return: Count of ST/SE loop errors
        @rtype: int
        """
        seg_err_ct = 0
        if self.child_err_count() > 0:
            seg_err_ct = 1
        return len(self.errors) + seg_err_ct

    def get_error_count(self):
        return self.err_count()

    def get_error_list(self, seg_id, pre=False):
        """
        """
        if seg_id == 'ST':
            return [err for err in self.errors if err[0] in ('1', '6', '7', '23')]
        elif seg_id == 'SE':
            return [err for err in self.errors if err[0] not in ('1', '6', '7', '23')]
        else:
            return []

    def child_err_count(self):
        ct = 0
        for child in self.children:
            if child.err_count() > 0:
                ct += 1
        return ct

    def get_cur_line(self):
        """
        @return: Current file line number
        @rtype: int
        """
        if self.cur_line_se:
            return self.cur_line_se
        else:
            return self.cur_line_st

    def is_closed(self):
        """
        @rtype: boolean
        """
        if self.cur_line_se:
            return True
        else:
            return False

    def __next__(self):
        """
        Return the next error node
        """
        for child in self.children:
            yield child
        return
        #self.parent.next()

    def __repr__(self):
        return '%i: %s' % (self.get_cur_line(), self.id)


class err_seg(err_node):
    """
    Segment Errors
    """
    def __init__(self, parent, map_node, seg_data, seg_count, cur_line, ls_id):
        """
        Needs:
            1. seg_id_code
            2. seg_pos - pos in ST loop
            3, loop_id - LS loop id
            4. seg_count - in parent
        """
        self.parent = parent
        if map_node is None:
            self.name = 'Unknown'
            self.pos = -1
        else:
            self.name = map_node.name
            self.pos = map_node.pos
        self.seg_id = seg_data.get_seg_id()
        self.seg_count = seg_count
        self.cur_line = cur_line
        self.ls_id = ls_id

        self.id = 'SEG'

        self.elements = []
        self.errors = []

    def accept(self, visitor):
        """
        Params:     visitor - ref to visitor class
        """
        visitor.visit_seg(self)
        for elem in self.elements:
            elem.accept(visitor)

    def add_error(self, err_cde, err_str, err_value=None):
        """
        @param err_cde: Error code
        @type err_cde: string
        @param err_str: Description of the error
        @type err_str: string
        """
        self.errors.append((err_cde, err_str, err_value))

    def err_count(self):
        """
        Returns:    count of errors
        """
        ele_err_ct = 0
        if self.child_err_count() > 0:
            ele_err_ct = 1
        return len(self.errors) + ele_err_ct

    def get_error_count(self):
        return self.err_count()

    def child_err_count(self):
        ct = 0
        for ele in self.elements:
            if ele.err_count() > 0:
                ct += 1
        return ct

    def __next__(self):
        """
        Desc:       Return the next error node
        """
        #for child in self.children:
        #    yield child
        #pdb.set_trace()
        return self
        #self.parent.next()

    def __repr__(self):
        return '%i: %s %s' % (self.get_cur_line(), self.id, self.seg_id)

    def get_first_child(self):
        return None
        #raise IterOutOfBounds


class err_ele(err_node):
    """
    Element Errors - Holds and generates output for element and
    composite/sub-element errors

    Each element with an error creates a new err_ele instance.
    """
    def __init__(self, parent, map_node):
        """
        """
        #, self.id, self.name, self.seq, self.data_ele)
        self.ele_ref_num = map_node.data_ele
        self.name = map_node.name
        if map_node.parent.is_composite():
            self.ele_pos = map_node.parent.seq
            self.subele_pos = map_node.seq
        else:
            self.ele_pos = map_node.seq
            self.subele_pos = None
        self.repeat_pos = None

        #self.bad_val = bad_val
        self.id = 'ELE'

        self.parent = parent
        #self.children = []
        self.errors = []

    def accept(self, visitor):
        """
        Params:     visitor - ref to visitor class
        """
        visitor.visit_ele(self)

    def add_error(self, err_cde, err_str, bad_value):
        """
        @param err_cde: Error code
        @type err_cde: string
        @param err_str: Description of the error
        @type err_str: string
        """
        #logger.debug('err_ele.add_error: %s %s %s' % (err_cde, err_str, bad_value))
        self.errors.append((err_cde, err_str, bad_value))

    def err_count(self):
        return len(self.errors)

    def get_error_count(self):
        return len(self.errors)


class ErrorErrhNull(Exception):
    """Class for errh_null errors."""


class errh_null(object):
    """
    A null error object - used for testing.
    Stores the current error in simple variables.
    """
    def __init__(self):
        self.id = 'ROOT'
        #self.children = []
        self.cur_node = self
        #self.cur_isa_node = None
        #self.cur_gs_node = None
        #self.cur_st_node = None
        #self.cur_seg_node = None
        #self.seg_node_added = False
        #self.cur_ele_node = None
        self.cur_line = 0
        self.err_cde = None
        self.err_str = None

    def reset(self):
        """
        Clear any errors
        """
        self.err_cde = None
        self.err_str = None

    def get_cur_line(self):
        """
        @return: Current file line number
        @rtype: int
        """
        return self.cur_line

    def get_id(self):
        """
        @return: Error node type
        @rtype: string
        """
        return self.id

    def add_isa_loop(self, seg, src):
        """
        """
        pass
        #raise ErrorErrhNull, 'add_isa loop'

    def add_gs_loop(self, seg, src):
        """
        """
        pass

    def add_st_loop(self, seg, src):
        """
        """
        pass

    def add_seg(self, map_node, seg, seg_count, cur_line, ls_id):
        """
        """
        pass

    def add_ele(self, map_node):
        """
        """
        pass

    def isa_error(self, err_cde, err_str):
        """
        @param err_cde: ISA level error code
        @type err_cde: string
        @param err_str: Description of the error
        @type err_str: string
        """
        self.err_cde = err_cde
        self.err_str = err_str

    def gs_error(self, err_cde, err_str):
        """
        @param err_cde: GS level error code
        @type err_cde: string
        @param err_str: Description of the error
        @type err_str: string
        """
        self.err_cde = err_cde
        self.err_str = err_str

    def st_error(self, err_cde, err_str):
        """
        @param err_cde: Segment level error code
        @type err_cde: string
        @param err_str: Description of the error
        @type err_str: string
        """
        self.err_cde = err_cde
        self.err_str = err_str

    def seg_error(self, err_cde, err_str, err_value=None, src_line=None):
        """
        @param err_cde: Segment level error code
        @type err_cde: string
        @param err_str: Description of the error
        @type err_str: string
        """
        self.err_cde = err_cde
        self.err_str = err_str

    def ele_error(self, err_cde, err_str, bad_value, refdes=None):
        """
        @param err_cde: Element level error code
        @type err_cde: string
        @param err_str: Description of the error
        @type err_str: string
        """
        self.err_cde = err_cde
        self.err_str = err_str

    def close_isa_loop(self, node, seg, src):
        """
        """
        pass

    def close_gs_loop(self, node, seg, src):
        """
        """
        pass

    def close_st_loop(self, node, seg, src):
        """
        """
        pass

    def find_node(self, type):
        """
        Find the last node of a type
        """
        pass

    def get_parent(self):
        return None

#    def get_first_child(self):
#        """
#        """
#        if len(self.children) > 0:
#            return self.children[0]
#        else:
#            return None

    def get_next_sibling(self):
        """
        """
        return None

    def get_error_count(self):
        """
        """
        if self.err_cde is not None:
            return 1
        else:
            return 0

    def handle_errors(self, err_list):
        pass

    def is_closed(self):
        """
        @rtype: boolean
        """
        return True

    def __repr__(self):
        """
        """
        return '%i: %s' % (-1, self.id)


class errh_list(object):
    """
    Capture validation errors in a list
    Used to refactor away from error_handler
    """
    def __init__(self):
        self.id = 'ROOT'
        self.cur_node = self
        self.cur_line = 0
        self.err_isa = []
        self.err_gs = []
        self.err_st = []
        self.err_seg = []
        self.err_ele = []

    def reset(self):
        """
        Clear any errors
        """
        self.err_isa = []
        self.err_gs = []
        self.err_st = []
        self.err_seg = []
        self.err_ele = []

    def get_cur_line(self):
        """
        @return: Current file line number
        @rtype: int
        """
        return self.cur_line

    def get_id(self):
        """
        @return: Error node type
        @rtype: string
        """
        return self.id

    def add_isa_loop(self, seg, src):
        pass

    def add_gs_loop(self, seg, src):
        pass

    def add_st_loop(self, seg, src):
        pass

    def add_seg(self, map_node, seg, seg_count, cur_line, ls_id):
        pass

    def add_ele(self, map_node):
        pass

    def isa_error(self, err_cde, err_str):
        """
        @param err_cde: ISA level error code
        @type err_cde: string
        @param err_str: Description of the error
        @type err_str: string
        """
        self.err_isa.append((err_cde, err_str))

    def gs_error(self, err_cde, err_str):
        """
        @param err_cde: GS level error code
        @type err_cde: string
        @param err_str: Description of the error
        @type err_str: string
        """
        self.err_gs.append((err_cde, err_str))

    def st_error(self, err_cde, err_str):
        """
        @param err_cde: Segment level error code
        @type err_cde: string
        @param err_str: Description of the error
        @type err_str: string
        """
        self.err_st.append((err_cde, err_str))

    def seg_error(self, err_cde, err_str, err_value=None, src_line=None):
        """
        @param err_cde: Segment level error code
        @type err_cde: string
        @param err_str: Description of the error
        @type err_str: string
        """
        self.err_seg.append((err_cde, err_str, err_value))

    def ele_error(self, err_cde, err_str, bad_value, refdes=None):
        """
        @param err_cde: Element level error code
        @type err_cde: string
        @param err_str: Description of the error
        @type err_str: string
        """
        self.err_ele.append((err_cde, err_str, bad_value, refdes))

    def close_isa_loop(self, node, seg, src):
        pass

    def close_gs_loop(self, node, seg, src):
        pass

    def close_st_loop(self, node, seg, src):
        pass

    def find_node(self, type):
        pass

    def get_parent(self):
        return None

    def get_next_sibling(self):
        return None

    def get_error_count(self):
        return len(self.err_isa) + len(self.err_gs) + len(self.err_st) + len(self.err_seg) + len(self.err_ele)

    def handle_errors(self, err_list):
        """
        Handles errors generated by X12Reader
        @param err_list: List of errors
        @type err_list: [(type, error code, error string)]
        """
        for (etype, err_cde, err_str, err_value, src_line) in err_list:
            if etype == 'isa':
                self.isa_error(err_cde, err_str)
            elif etype == 'gs':
                self.gs_error(err_cde, err_str)
            elif etype == 'st':
                self.st_error(err_cde, err_str)
            elif etype == 'seg':
                self.seg_error(err_cde, err_str, err_value, src_line)

    def is_closed(self):
        """
        @rtype: boolean
        """
        return True

    def __repr__(self):
        """
        """
        return '%i: %s' % (-1, self.id)



######################################################################
# Copyright 
#   John Holland <john@zoner.org>
# All rights reserved.
#
# This software is licensed as described in the file LICENSE.txt, which
# you should have received as part of this distribution.
#
######################################################################

"""
X12 data element validation
"""
import re

# Intrapackage imports
from errors import IsValidError, EngineError


def IsValidDataType(str_val, data_type, charset='B', icvn='00401'):
    """
    Is str_val a valid X12 data value

    @param str_val: data value to validate
    @type str_val: string
    @param data_type: X12 data element identifier
    @type data_type: string
    @param charset: [optional] - 'B' for Basic X12 character set, 'E' for extended
    @type charset: string
    @rtype: boolean
    @todo: need to generalize control character validation
    """
    if not data_type:
        return True
    if not isinstance(str_val, str):
        return False

    try:
        if data_type[0] == 'N':
            if not match_re('N', str_val):
                raise IsValidError  # not a number
        elif data_type == 'R':
            if not match_re('R', str_val):
                raise IsValidError  # not a number
        elif data_type in ('ID', 'AN'):
            if not_match_re('ID', str_val, charset, icvn):
                raise IsValidError
        elif data_type == 'RD8':
            if '-' in str_val:
                (start, end) = str_val.split('-')
                return IsValidDataType(start, 'D8', charset) and IsValidDataType(end, 'D8', charset)
            else:
                return False
        elif data_type in ('DT', 'D8', 'D6'):
            if not is_valid_date(data_type, str_val):
                raise IsValidError
        elif data_type == 'TM':
            if not is_valid_time(str_val):
                raise IsValidError
        elif data_type == 'B':
            pass
        else:
            raise IsValidError('Unknown data type %s' % data_type)
    except IsValidError:
        return False
    return True

rec_N = re.compile("^-?[0-9]+", re.S)
rec_R = re.compile("^-?[0-9]*(\.[0-9]+)?", re.S)
rec_ID_E = re.compile(
    "[^A-Z0-9!\"&'()*+,\-\./:;?=\sa-z%~@\[\]_{}\\\|<>#$\s]", re.S)
rec_ID_E5 = re.compile(
    "[^A-Z0-9!\"&'()*+,\-\./:;?=\sa-z%~@\[\]_{}\\\|<>^`#$\s]", re.S)
rec_ID_B = re.compile("[^A-Z0-9!\"&'()*+,\-\./:;?=\s]", re.S)
rec_DT = re.compile("[^0-9]+", re.S)
rec_TM = re.compile("[^0-9]+", re.S)


def match_re(short_data_type, val):
    """
    @param short_data_type: simplified data type
    @type short_data_type: string
    @param val: data value to be verified
    @type val: string
    @return: True if matched, False if not
    @rtype: boolean
    """
    if short_data_type == 'N':
        rec = rec_N
    elif short_data_type == 'R':
        rec = rec_R
    else:
        raise EngineError('Unknown data type %s' % (short_data_type))
    m = rec.search(val)
    if not m:
        return False
    if m.group(0) != val:  # matched substring != original, bad
        return False  # nothing matched
    return True


def not_match_re(short_data_type, val, charset='B', icvn='00401'):
    """
    @param short_data_type: simplified data type
    @type short_data_type: string
    @param val: data value to be verified
    @type val: string
    @param charset: [optional] - 'B' for Basic X12 character set, 'E' for extended, E5 for 5010 Extended
    @type charset: string
    @return: True if found invalid characters, False if none
    @rtype: boolean
    """
    if short_data_type in ('ID', 'AN'):
        if charset == 'E':  # extended charset
            if icvn == '00501':
                rec = rec_ID_E5
            else:
                rec = rec_ID_E
        elif charset == 'B':  # basic charset:
            rec = rec_ID_B
    elif short_data_type == 'DT':
        rec = rec_DT
    elif short_data_type == 'TM':
        rec = rec_TM
    else:
        raise EngineError('Unknown data type %s' % (short_data_type))
    m = rec.search(val)
    if m and m.group(0):
        return True  # Invalid char matched
    return False


def is_valid_date(data_type, val):
    """
    @param data_type: Date type
    @type data_type: string
    @param val: data value to be verified
    @type val: string
    @return: True if valid, False if not
    @rtype: boolean
    """
    try:
        if data_type == 'D8' and len(val) != 8:
            raise IsValidError
        if data_type == 'D6' and len(val) != 6:
            raise IsValidError
        if not_match_re('DT', val):
            raise IsValidError
        if len(val) in (6, 8, 12):  # valid lengths for date
            try:
                if 6 == len(val):  # if 2 digit year, add CC
                    val = '20' + val if val[0:2] < 50 else '19' + val
                year = int(val[0:4])  # get year
                month = int(val[4:6])
                day = int(val[6:8])
                # Should not have dates before 1/1/1800
                if year < 1800:
                    raise IsValidError
                # check month
                if month < 1 or month > 12:
                    raise IsValidError
                if month in (1, 3, 5, 7, 8, 10, 12):  # 31 day month
                    if day < 1 or day > 31:
                        raise IsValidError
                elif month in (4, 6, 9, 11):  # 30 day month
                    if day < 1 or day > 30:
                        raise IsValidError
                else:  # else 28 day
                    if not year % 4 and not (not year % 100 and year % 400):
                    #if not (year % 4) and ((year % 100) or (not (year % 400)) ):  # leap year
                        if day < 1 or day > 29:
                            raise IsValidError
                    elif day < 1 or day > 28:
                        raise IsValidError
                if len(val) == 12:
                    if not is_valid_time(val[8:12]):
                        raise IsValidError
            except TypeError:
                raise IsValidError
        else:
            raise IsValidError
    except IsValidError:
        return False
    return True


def is_valid_time(val):
    """
    @param val: time value to be verified
    @type val: string
    """
    try:
        not_match_re('TM', val)
        if val[0:2] > '23' or val[2:4] > '59':  # check hour, minute segment
            raise IsValidError
        elif len(val) > 4:  # time contains seconds
            if len(val) < 6:  # length is munted
                raise IsValidError
            elif val[4:6] > '59':  # check seconds
                raise IsValidError
            # check decimal seconds here in the future
            elif len(val) > 8:
                # print 'unhandled decimal seconds encountered'
                raise IsValidError
    except IsValidError:
        return False
    return True


def contains_control_character(str_val, charset='B', icvn='00401'):
    if '\n' in str_val:
        return (True, '<LF>')
    if '\r' in str_val:
        return (True, '<CR>')
    return (False, None)



from setuptools import setup


long_description = """Pyx12 is a HIPAA X12 document validator and converter. It parses an ANSI X12N data file and
validates it against a representation of the Implementation Guidelines for a HIPAA transaction. By default, it
creates a 997 response for 4010 and a 999 response for 5010. It can create an html representation of the X12
document or can translate to and from an XML representation of the data file."""

setup(
    name="pyx12",
    version="2.2.0",
    long_description=long_description,
    license='BSD',
    description="HIPAA X12 validator, parser and converter",
    keywords='x12 hipaa healthcare edi',
    author="John Holland",
    author_email="john.holland@swmbh.org",
    url="http://github.com/azoner/pyx12",
    platforms='All',
    packages=['pyx12', 'pyx12.scripts'],
    package_data={
        '': ['*.xml', '*.md'],
        'pyx12': ['map/*.xml', 'map/*.xsd'],
    },
    #data_files=[('config', ['bin/pyx12.conf.xml.sample'])],
    entry_points={
        'console_scripts': [
            'x12html = pyx12.scripts.x12html:main',
            'x12valid = pyx12.scripts.x12valid:main',
            'x12info = pyx12.scripts.x12info:main',
            'x12norm = pyx12.scripts.x12norm:main',
            'x12xml = pyx12.scripts.x12xml:main',
            'xmlx12 = pyx12.scripts.xmlx12:main',
        ]
    },
    test_suite="pyx12.test",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Healthcare Industry',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Topic :: Office/Business',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)
